"""Generate tables for benchmarking PostgreSQL backend."""

from itertools import chain
from pathlib import Path
from typing import Iterable

import click
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq

from matchbox.common.factories.models import generate_dummy_probabilities
from matchbox.common.hash import HASH_FUNC, hash_data, hash_values
from matchbox.common.logging import console
from matchbox.common.transform import (
    attach_components_to_probabilities,
    to_hierarchical_clusters,
)
from matchbox.server.postgresql.utils.insert import HashIDMap

PRESETS = {
    "xs": {
        "source_len": 10_000,
        "dedupe_components": 8000,
        "dedupe_len": 2000,
        "link_components": 6000,
        "link_len": 10_000,
    },
    "s": {
        "source_len": 100_000,
        "dedupe_components": 80_000,
        "dedupe_len": 20_000,
        "link_components": 60_000,
        "link_len": 100_000,
    },
    "m": {
        "source_len": 1_000_000,
        "dedupe_components": 800_000,
        "dedupe_len": 200_000,
        "link_components": 600_000,
        "link_len": 1_000_000,
    },
    "l": {
        "source_len": 10_000_000,
        "dedupe_components": 8_000_000,
        "dedupe_len": 2_000_000,
        "link_components": 6_000_000,
        "link_len": 10_000_000,
    },
    "xl": {
        "source_len": 100_000_000,
        "dedupe_components": 80_000_000,
        "dedupe_len": 20_000_000,
        "link_components": 60_000_000,
        "link_len": 100_000_000,
    },
}


def _hash_list_int(li: list[int]) -> list[bytes]:
    return [HASH_FUNC(str(i).encode("utf-8")).digest() for i in li]


def generate_sources(
    dataset_start_id: int = 1,
) -> tuple[pa.Table, pa.Table]:
    """Generate sources and source_columns tables.

    Args:
        dataset_start_id: Starting ID for dataset resolution IDs

    Returns:
        Tuple of (sources_table, source_columns_table) as PyArrow tables
    """
    # Sources data
    sources_id = [dataset_start_id, dataset_start_id + 1]
    sources_resolution_id = sources_id
    sources_resolution_names = ["source1@warehouse", "source2@warehouse"]
    sources_full_names = ["dbt.companies_house", "dbt.hmrc_exporters"]
    sources_db_pk = ["company_number", "id"]
    warehouse_hashes = [bytes("foo".encode("ascii"))] * 2

    # Create sources table without column arrays
    sources_table = pa.table(
        {
            "source_id": pa.array(sources_id, type=pa.uint64()),
            "resolution_id": pa.array(sources_resolution_id, type=pa.uint64()),
            "resolution_name": pa.array(sources_resolution_names, type=pa.string()),
            "full_name": pa.array(sources_full_names, type=pa.string()),
            "warehouse_hash": pa.array(warehouse_hashes, type=pa.large_binary()),
            "db_pk": pa.array(sources_db_pk, type=pa.string()),
        }
    )

    # Column data
    column_names = [["col1"], ["col2"]]
    column_types = [["TEXT"], ["TEXT"]]

    # Create flattened data for source_columns table
    column_ids = []
    source_ids = []
    column_indices = []
    column_name_values = []
    column_type_values = []

    # Start with column_id = 1
    next_column_id = 1

    # Process each source and its columns
    for i, source_id in enumerate(sources_id):
        for j, (name, type_val) in enumerate(
            zip(column_names[i], column_types[i], strict=True)
        ):
            column_ids.append(next_column_id)
            next_column_id += 1
            source_ids.append(source_id)
            column_indices.append(j)
            column_name_values.append(name)
            column_type_values.append(type_val)

    # Create source_columns table
    source_columns_table = pa.table(
        {
            "column_id": pa.array(column_ids, type=pa.uint64()),
            "source_id": pa.array(source_ids, type=pa.uint64()),
            "column_index": pa.array(column_indices, type=pa.int32()),
            "column_name": pa.array(column_name_values, type=pa.string()),
            "column_type": pa.array(column_type_values, type=pa.string()),
        }
    )

    return sources_table, source_columns_table


def generate_resolutions(dataset_start_id: int = 1) -> pa.Table:
    """Generate resolutions table.

    Args:
        dataset_start_id: Starting ID for dataset resolution IDs

    Returns:
        PyArrow resolutions table
    """
    base_id = dataset_start_id
    resolutions_resolution_id = [
        base_id,
        base_id + 1,
        base_id + 2,
        base_id + 3,
        base_id + 4,
    ]
    resolutions_name = ["source1", "source2", "dedupe1", "dedupe2", "link"]
    resolutions_resolution_hash = [
        HASH_FUNC(rid.encode("utf-8")).digest() for rid in resolutions_name
    ]
    resolutions_type = ["dataset", "dataset", "model", "model", "model"]
    resolutions_truth = [None, None, 80, 80, 90]

    return pa.table(
        {
            "resolution_id": pa.array(resolutions_resolution_id, type=pa.uint64()),
            "resolution_hash": pa.array(
                resolutions_resolution_hash, type=pa.large_binary()
            ),
            "type": pa.array(resolutions_type, type=pa.string()),
            "name": pa.array(resolutions_name, type=pa.string()),
            "description": pa.array(resolutions_name, type=pa.string()),
            "truth": pa.array(resolutions_truth, type=pa.int8()),
        }
    )


def generate_resolution_from(dataset_start_id: int = 1) -> pa.Table:
    """Generate resolution_from table.

    Args:
        dataset_start_id: Starting ID for dataset resolution IDs

    Returns:
        PyArrow resolution_from table
    """
    base_id = dataset_start_id
    # 1 and 2 are sources; 3 and 4 are dedupers; 5 is a linker
    resolution_parent = [
        base_id,
        base_id,
        base_id + 2,
        base_id + 1,
        base_id + 1,
        base_id + 3,
    ]
    resolution_child = [
        base_id + 2,
        base_id + 4,
        base_id + 4,
        base_id + 3,
        base_id + 4,
        base_id + 4,
    ]
    resolution_level = [1, 2, 1, 1, 2, 1]
    resolution_truth_cache = [None, None, 70, None, None, 70]

    return pa.table(
        {
            "parent": pa.array(resolution_parent, type=pa.uint64()),
            "child": pa.array(resolution_child, type=pa.uint64()),
            "level": pa.array(resolution_level, type=pa.int32()),
            "truth_cache": pa.array(resolution_truth_cache, type=pa.uint8()),
        }
    )


def generate_cluster_source(
    range_left: int,
    range_right: int,
    source_id: int,
    cluster_start_id: int = 0,
    pk_start_id: int = 0,  # Need to track the global pk_id to ensure uniqueness
) -> tuple[pa.Table, pa.Table]:
    """Generate both Clusters and ClusterSourcePK tables for source rows.

    Args:
        range_left: first ID to generate
        range_right: last ID to generate, plus one
        source_id: source ID for the source
        cluster_start_id: Starting ID for clusters
        pk_start_id: Starting ID for primary keys

    Returns:
        Tuple of (Clusters table, ClusterSourcePK table)
    """
    # Generate cluster IDs
    source = list(range(cluster_start_id + range_left, cluster_start_id + range_right))

    # Create the Clusters table (without dataset column)
    clusters_table = pa.table(
        {
            "cluster_id": pa.array(source, type=pa.uint64()),
            "cluster_hash": pa.array(_hash_list_int(source), type=pa.large_binary()),
        }
    )

    # Generate data and create ClusterSourcePK table
    pk_ids = []
    cluster_ids = []
    source_ids = []
    source_pks = []

    for i, cluster_id in enumerate(source):
        # Create a single source_pk entry for each cluster
        pk_ids.append(pk_start_id + i)
        cluster_ids.append(cluster_id)
        source_ids.append(source_id)
        source_pks.append(str(cluster_id))

    cluster_source_pks_table = pa.table(
        {
            "pk_id": pa.array(pk_ids, type=pa.int64()),
            "cluster_id": pa.array(cluster_ids, type=pa.uint64()),
            "source_id": pa.array(source_ids, type=pa.uint64()),
            "source_pk": pa.array(source_pks, type=pa.string()),
        }
    )

    return clusters_table, cluster_source_pks_table


def generate_result_tables(
    left_ids: Iterable[int],
    right_ids: Iterable[int] | None,
    resolution_id: int,
    next_id: int,
    n_components: int,
    n_probs: int,
    prob_min: float = 0.6,
    prob_max: float = 1,
) -> tuple[list[int], pa.Table, pa.Table, pa.Table, int]:
    """Generate probabilities, contains and clusters tables.

    Args:
        left_ids: list of IDs for rows to dedupe, or for left rows to link
        right_ids: list of IDs for right rows to link
        resolution_id: ID of resolution for this dedupe or link model
        next_id: the next ID to use when generating IDs
        n_components: number of implied connected components
        n_probs: total number of probability edges to be generated
        prob_min: minimum value for probabilities to be generated
        prob_max: maximum value for probabilities to be generated

    Returns:
        Tuple with 1 list of top-level clusters, 3 PyArrow tables, for probabilities,
        contains, and clusters, and the next ID to use for future calls
    """
    probs = generate_dummy_probabilities(
        tuple(left_ids),
        tuple(right_ids) if right_ids else None,
        (prob_min, prob_max),
        n_components,
        n_probs,
    )

    # Create a lookup table for hashes
    all_probs = pa.concat_arrays(
        [probs["left_id"].combine_chunks(), probs["right_id"].combine_chunks()]
    )

    lookup = pa.table(
        {
            "id": all_probs,
            "hash": pa.array(
                [hash_data(p) for p in all_probs.to_pylist()], type=pa.large_binary()
            ),
        }
    )

    hm = HashIDMap(start=next_id, lookup=lookup)

    # Join hashes, probabilities and components
    probs_with_ccs = attach_components_to_probabilities(
        pa.table(
            {
                "left_id": hm.get_hashes(probs["left_id"]),
                "right_id": hm.get_hashes(probs["right_id"]),
                "probability": probs["probability"],
            }
        )
    )

    # Calculate hierarchies
    hierarchy = to_hierarchical_clusters(
        probabilities=probs_with_ccs,
        hash_func=hash_values,
        dtype=pa.large_binary,
    )

    # Shape into tables
    new_hierarchy_schema = pa.schema(
        [
            pa.field("parent", pa.uint64()),
            pa.field("child", pa.uint64()),
            pa.field("probability", pa.uint8()),
        ]
    )
    hierarchy = pa.Table.from_arrays(
        [
            pa.array(hm.get_ids(hierarchy["parent"]), type=pa.uint64()),
            pa.array(hm.get_ids(hierarchy["child"]), type=pa.uint64()),
            hierarchy["probability"],
        ],
        schema=new_hierarchy_schema,
    ).sort_by([("parent", "ascending")])

    prev_parent = None
    unique_indices = []
    i = -1
    for batch in hierarchy.to_batches():
        d = batch.to_pydict()
        for parent in d["parent"]:
            i += 1
            if parent != prev_parent:
                unique_indices.append(i)
                prev_parent = parent

    mask = np.full((len(hierarchy)), False)
    mask[unique_indices] = True
    unique_parent_hierarchy = hierarchy.filter(mask=mask)
    unique_parent_ids = unique_parent_hierarchy["parent"].combine_chunks()

    parent_ids = hierarchy["parent"]
    child_ids = hierarchy["child"]
    unique_child_ids = pc.unique(child_ids)

    probabilities_table = pa.table(
        {
            "resolution": pa.array(
                [resolution_id] * len(unique_parent_ids), type=pa.uint64()
            ),
            "cluster": unique_parent_ids,
            "probability": unique_parent_hierarchy["probability"],
        }
    )

    contains_table = pa.table(
        {
            "parent": parent_ids,
            "child": child_ids,
        }
    )

    clusters_table = pa.table(
        {
            "cluster_id": unique_parent_ids,
            "cluster_hash": hm.get_hashes(unique_parent_ids),
        }
    )

    # Compute top clusters
    parents_not_children = pc.filter(
        unique_parent_ids, pc.invert(pc.is_in(unique_parent_ids, unique_child_ids))
    )
    right_ids_or_empty = [] if not right_ids else right_ids
    all_sources = pa.array(chain(left_ids, right_ids_or_empty), type=pa.uint64())

    sources_not_children = pc.filter(
        all_sources, pc.invert(pc.is_in(all_sources, unique_child_ids))
    )
    top_clusters = pc.unique(
        pa.concat_arrays([parents_not_children, sources_not_children])
    )

    return (
        top_clusters,
        probabilities_table,
        contains_table,
        clusters_table,
        hm.next_int,
    )


def generate_all_tables(
    source_len: int,
    dedupe_components: int,
    dedupe_len: int,
    link_components: int,
    link_len: int,
    cluster_start_id: int = 0,
    dataset_start_id: int = 1,
    pk_start_id: int = 0,
) -> dict[str, pa.Table]:
    """Make all six PostgreSQL backend tables.

    It will create two sources, one deduper for each, and one linker from
    each deduper.

    Args:
        source_len: length of each data source
        dedupe_components: number of connected components implied by each deduper
        dedupe_len: probabilities generated by each deduper
        link_components: number of connected components implied by each linker
        link_len: probabilities generated by each linker
        cluster_start_id: Starting ID for clusters
        dataset_start_id: Starting ID for dataset resolution IDs
        pk_start_id: Starting ID for primary keys (globally unique)

    Returns:
        A dictionary where keys are table names and values are PyArrow tables
    """
    console.log("Generating sources")
    resolutions = generate_resolutions(dataset_start_id)
    resolution_from = generate_resolution_from(dataset_start_id)
    sources, columns = generate_sources(dataset_start_id)

    clusters_source1, source_pks1 = generate_cluster_source(
        range_left=0,
        range_right=source_len,
        source_id=dataset_start_id,
        cluster_start_id=cluster_start_id,
        pk_start_id=pk_start_id,
    )

    current_pk_id = pk_start_id + len(source_pks1)

    clusters_source2, source_pks2 = generate_cluster_source(
        range_left=source_len,
        range_right=source_len * 2,
        source_id=dataset_start_id + 1,
        cluster_start_id=cluster_start_id,
        pk_start_id=current_pk_id,
    )

    current_pk_id += len(source_pks2)

    initial_next_id = cluster_start_id + (source_len * 2)

    console.log("Generating the deduplication tables")
    (
        top_clusters1,
        probabilities_dedupe1,
        contains_dedupe1,
        clusters_dedupe1,
        next_id1,
    ) = generate_result_tables(
        left_ids=clusters_source1["cluster_id"].to_pylist(),
        right_ids=None,
        resolution_id=dataset_start_id + 2,
        next_id=initial_next_id,
        n_components=dedupe_components,
        n_probs=dedupe_len,
    )

    (
        top_clusters2,
        probabilities_dedupe2,
        contains_dedupe2,
        clusters_dedupe2,
        next_id2,
    ) = generate_result_tables(
        left_ids=clusters_source2["cluster_id"].to_pylist(),
        right_ids=None,
        resolution_id=dataset_start_id + 3,
        next_id=next_id1,
        n_components=dedupe_components,
        n_probs=dedupe_len,
    )

    console.log("Generating the link tables")
    _, probabilities_link, contains_link, clusters_link, final_next_id = (
        generate_result_tables(
            left_ids=top_clusters1,
            right_ids=top_clusters2,
            resolution_id=dataset_start_id + 4,
            next_id=next_id2,
            n_components=link_components,
            n_probs=link_len,
        )
    )

    probabilities = pa.concat_tables(
        [probabilities_dedupe1, probabilities_dedupe2, probabilities_link]
    ).combine_chunks()
    contains = pa.concat_tables(
        [contains_dedupe1, contains_dedupe2, contains_link]
    ).combine_chunks()
    clusters = pa.concat_tables(
        [
            clusters_source1,
            clusters_source2,
            clusters_dedupe1,
            clusters_dedupe2,
            clusters_link,
        ]
    ).combine_chunks()

    # Combine the source primary keys tables
    cluster_source_pks = pa.concat_tables([source_pks1, source_pks2]).combine_chunks()

    console.log("Generation complete.")
    console.log(f"Next dataset id: {dataset_start_id + 5}")
    console.log(f"Next cluster id: {final_next_id}")
    console.log(f"Next pk id: {current_pk_id}")

    return {
        "resolutions": resolutions,
        "resolution_from": resolution_from,
        "sources": sources,
        "source_columns": columns,
        "clusters": clusters,
        "cluster_source_pks": cluster_source_pks,
        "contains": contains,
        "probabilities": probabilities,
    }


@click.command()
@click.option("-s", "--settings", type=str, required=True, help="Settings dict to use.")
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(exists=True, path_type=Path),
    help="Directory to save parquets to.",
)
@click.option(
    "-c", "--cluster-start-id", type=int, default=0, help="Starting ID for clusters"
)
@click.option(
    "-d",
    "--dataset-start-id",
    type=int,
    default=1,
    help="Starting ID for dataset resolution IDs",
)
def main(
    settings: str,
    output_dir: Path,
    cluster_start_id: int,
    dataset_start_id: int,
) -> None:
    """Command line tool for generating data.

    Args:
        settings: The key of the settings dictionary to use
        output_dir: Where to save the output files
        cluster_start_id: The first integer to use for clusters
        dataset_start_id: The first integer to use for datasets

    Examples:
        ```shell
        generate_tables.py \
            --settings xl \
            --output-dir data/v4 \
            --dataset-start-id 6742 \
            --cluster-start-id 7
        ```
        ```shell
        generate_tables.py -s s -o data/v4 -d 1 -c 0
        ```
    """
    if not output_dir:
        output_dir = Path.cwd() / "data" / "all_tables"
    if settings not in PRESETS:
        raise ValueError(f"Settings {settings} are invalid")

    config = PRESETS[settings]
    source_len = config["source_len"]
    dedupe_components = config["dedupe_components"]
    dedupe_len = config["dedupe_len"]
    link_len = config["link_len"]
    link_components = config["link_components"]

    all_tables = generate_all_tables(
        source_len=source_len,
        dedupe_components=dedupe_components,
        dedupe_len=dedupe_len,
        link_components=link_components,
        link_len=link_len,
        cluster_start_id=cluster_start_id,
        dataset_start_id=dataset_start_id,
    )

    output_dir /= settings
    output_dir.mkdir(parents=True, exist_ok=True)

    console.log("Writing to disk")
    for name, table in all_tables.items():
        pq.write_table(table, output_dir / f"{name}.parquet")

    console.log("Complete")


if __name__ == "__main__":
    main()



"""General utilities for the PostgreSQL backend."""

import base64
import contextlib
import cProfile
import io
import pstats
import uuid

import pyarrow as pa
from adbc_driver_manager import ProgrammingError as ADBCProgrammingError
from adbc_driver_postgresql import dbapi as adbc_dbapi
from pyarrow import Table as ArrowTable
from sqlalchemy import Column, Engine, Table, inspect, select
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import DeclarativeMeta, Session
from sqlalchemy.sql import Select

from matchbox.common.exceptions import (
    MatchboxDatabaseWriteError,
    MatchboxResolutionNotFoundError,
)
from matchbox.common.graph import (
    ResolutionEdge,
    ResolutionGraph,
    ResolutionNode,
    ResolutionNodeType,
)
from matchbox.server.base import MatchboxBackends, MatchboxSnapshot
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourcePK,
    Contains,
    Probabilities,
    ResolutionFrom,
    Resolutions,
    SourceColumns,
    Sources,
)

# Retrieval


def resolve_model_name(model: str, engine: Engine) -> Resolutions:
    """Resolves a model name to a Resolution object.

    Args:
        model: The name of the model to resolve.
        engine: The database engine.

    Raises:
        MatchboxResolutionNotFoundError: If the model doesn't exist.
    """
    with Session(engine) as session:
        if (
            resolution := session.query(Resolutions)
            .filter_by(name=model, type="model")
            .first()
        ):
            return resolution
        raise MatchboxResolutionNotFoundError(
            message=f"Resolution {model} not found or not of type 'model'."
        )


def get_resolution_graph(engine: Engine) -> ResolutionGraph:
    """Retrieves the resolution graph."""
    G = ResolutionGraph(nodes=set(), edges=set())
    with Session(engine) as session:
        for resolution in session.query(Resolutions).all():
            G.nodes.add(
                ResolutionNode(
                    id=resolution.resolution_id,
                    name=resolution.name,
                    type=ResolutionNodeType(resolution.type),
                )
            )

        for edge in (
            session.query(ResolutionFrom).filter(ResolutionFrom.level == 1).all()
        ):
            G.edges.add(ResolutionEdge(parent=edge.parent, child=edge.child))

    return G


# Data management


def dump(engine: Engine) -> MatchboxSnapshot:
    """Dumps the entire database to a snapshot.

    Args:
        engine: The database engine.

    Returns:
        A MatchboxSnapshot object of type "postgres" with the database's
            current state.
    """
    tables = {
        "resolutions": Resolutions,
        "resolution_from": ResolutionFrom,
        "sources": Sources,
        "source_columns": SourceColumns,
        "clusters": Clusters,
        "cluster_source_pks": ClusterSourcePK,
        "contains": Contains,
        "probabilities": Probabilities,
    }

    data = {}

    with Session(engine) as session:
        for table_name, model in tables.items():
            # Query all records from this table
            records = session.execute(select(model)).scalars().all()

            # Convert each record to a dictionary
            table_data = []
            for record in records:
                record_dict = {}
                for column in inspect(model).columns:
                    value = getattr(record, column.name)

                    # Store bytes as nested dictionary with encoding format
                    if isinstance(value, bytes):
                        record_dict[column.name] = {
                            "base64": base64.b64encode(value).decode("ascii")
                        }
                    else:
                        record_dict[column.name] = value

                table_data.append(record_dict)

            data[table_name] = table_data

    return MatchboxSnapshot(backend_type=MatchboxBackends.POSTGRES, data=data)


def restore(engine: Engine, snapshot: MatchboxSnapshot, batch_size: int) -> None:
    """Restores the database from a snapshot.

    Args:
        engine: The database engine.
        snapshot: A MatchboxSnapshot object of type "postgres" with the
            database's state
        batch_size: The number of records to insert in each batch

    Raises:
        ValueError: If the snapshot is missing data
    """
    table_map = {
        "resolutions": Resolutions,
        "resolution_from": ResolutionFrom,
        "sources": Sources,
        "source_columns": SourceColumns,
        "clusters": Clusters,
        "cluster_source_pks": ClusterSourcePK,
        "contains": Contains,
        "probabilities": Probabilities,
    }

    table_order = [
        "resolutions",
        "resolution_from",
        "sources",
        "source_columns",
        "clusters",
        "cluster_source_pks",
        "contains",
        "probabilities",
    ]

    with Session(engine) as session:
        # Process tables in order
        for table_name in table_order:
            if table_name not in snapshot.data:
                raise ValueError(f"Invalid: Table {table_name} not found in snapshot.")

            model = table_map[table_name]
            records = snapshot.data[table_name]

            if not records:
                continue

            # Process records for insertion
            processed_records = []
            for record in records:
                processed_record = {}

                for key, value in record.items():
                    # Check if the value is a dictionary with encoding format
                    if isinstance(value, dict) and "base64" in value:
                        processed_record[key] = base64.b64decode(value["base64"])
                    else:
                        processed_record[key] = value

                processed_records.append(processed_record)

            # Insert
            for i in range(0, len(processed_records), batch_size):
                batch = processed_records[i : i + batch_size]
                session.bulk_insert_mappings(model, batch)
                session.flush()

        session.commit()


# SQLAlchemy profiling


@contextlib.contextmanager
def sqa_profiled():
    """SQLAlchemy profiler.

    Taken directly from their docs:
    https://docs.sqlalchemy.org/en/20/faq/performance.html#query-profiling
    """
    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats("cumulative")
    ps.print_stats()
    # uncomment this to see who's calling what
    # ps.print_callers()
    print(s.getvalue())


# Misc


def compile_sql(stmt: Select) -> str:
    """Compiles a SQLAlchemy statement into a string.

    Args:
        stmt: The SQLAlchemy statement to compile.

    Returns:
        The compiled SQL statement as a string.
    """
    return str(
        stmt.compile(
            dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}
        )
    )


def _copy_to_table(
    table_name: str,
    schema_name: str,
    data: ArrowTable,
    connection: adbc_dbapi.Connection,
    max_chunksize: int | None = None,
):
    batch_reader = pa.RecordBatchReader.from_batches(
        data.schema, data.to_batches(max_chunksize=max_chunksize)
    )

    with connection.cursor() as cursor:
        cursor.adbc_ingest(
            table_name=table_name,
            data=batch_reader,
            mode="append",
            db_schema_name=schema_name,
        )
        connection.commit()


def large_ingest(
    data: pa.Table,
    table_class: DeclarativeMeta,
    max_chunksize: int | None = None,
    upsert_keys: list[str] | None = None,
    update_columns: list[str] | None = None,
):
    """Append a PyArrow table to a PostgreSQL table using ADBC.

    It will either copy directly (and error if primary key constraints are violated),
    or it can be run in upsert mode by using a staging table, which is slower.

    Args:
        data: A PyArrow table to write.
        table_class: The SQLAlchemy ORM class for the table to write to.
        max_chunksize: Size of data chunks to be read and copied.
        upsert_keys: Columns used as keys for "on conflict do update".
            If passed, it will run ingest in slower upsert mode.
            If not passed and `update_columns` is passed, defaults to primary keys.
        update_columns: Columns to update when upserting.
            If passed, it will run ingest in slower upsert mode.
            If not passed and `upsert_keys` is passed, defaults to all other columns.
    """
    with (
        MBDB.get_adbc_connection() as conn,
        MBDB.get_session() as session,
    ):
        table: Table = table_class.__table__
        metadata = table.metadata

        table_columns = [c.name for c in table.columns]
        col_diff = set(data.column_names) - set(table_columns)
        if len(col_diff) > 0:
            raise ValueError(f"Table {table.name} does not have columns {col_diff}")

        if not update_columns and not upsert_keys:
            try:
                _copy_to_table(
                    table_name=table.name,
                    schema_name=table.schema,
                    data=data,
                    connection=conn,
                    max_chunksize=max_chunksize,
                )
            except ADBCProgrammingError as e:
                raise MatchboxDatabaseWriteError from e

        # Upsert mode (slower)
        else:
            pk_names = [c.name for c in table.primary_key.columns]

            # Validate upsert arguments
            if len(set(update_columns or []) & set(upsert_keys or [])) > 0:
                raise ValueError("Cannot update a custom upsert key")

            if len(set(update_columns or []) & set(pk_names)) > 0:
                raise ValueError(
                    "Cannot update a primary key without "
                    "setting a different custom upsert key"
                )

            # If necessary, set defaults for upsert variables
            upsert_keys = upsert_keys or pk_names

            if not update_columns:
                update_columns = [c for c in table_columns if c not in upsert_keys]
            try:
                # Create temp table
                temp_table_name = f"{table.name}_tmp_{uuid.uuid4().hex}"
                temp_cols = [
                    Column(c.name, c.type, primary_key=c.primary_key)
                    for c in table.columns
                ]
                temp_table = Table(temp_table_name, metadata, *temp_cols)
                temp_table.create(session.bind)

                # Add new records to temp table
                _copy_to_table(
                    table_name=temp_table_name,
                    schema_name=table.schema,
                    data=data,
                    connection=conn,
                    max_chunksize=max_chunksize,
                )

                # Copy new records to original table
                insert_stmt = insert(table).from_select(
                    [c.name for c in temp_table.columns], temp_table.select()
                )
                upsert_stmt = insert_stmt.on_conflict_do_update(
                    index_elements=upsert_keys,
                    set_={c: getattr(insert_stmt.excluded, c) for c in update_columns},
                )

                session.execute(upsert_stmt)
                session.commit()

            except (ADBCProgrammingError, IntegrityError) as e:
                raise MatchboxDatabaseWriteError from e

            finally:
                # Drop temp table
                temp_table.drop(session.bind, checkfirst=True)



import logging
from functools import partial

import pytest
from httpx import Client
from pandas import DataFrame
from sqlalchemy import Engine, text

from matchbox import index, make_model, process, query
from matchbox.client.clean import steps
from matchbox.client.clean.utils import cleaning_function
from matchbox.client.helpers import cleaner, cleaners, select
from matchbox.client.models.dedupers import NaiveDeduper
from matchbox.client.models.linkers import DeterministicLinker
from matchbox.common.db import fullname_to_prefix
from matchbox.common.factories.entities import query_to_cluster_entities
from matchbox.common.factories.sources import (
    FeatureConfig,
    LinkedSourcesTestkit,
    SourceConfig,
    SuffixRule,
    linked_sources_factory,
)


@pytest.mark.docker
class TestE2EAnalyticalUser:
    """End to end tests for analytical user functionality."""

    client: Client | None = None
    warehouse_engine: Engine | None = None
    linked_testkit: LinkedSourcesTestkit | None = None
    n_true_entities: int | None = None

    @pytest.fixture(autouse=True, scope="function")
    def setup_environment(
        self,
        matchbox_client: Client,
        postgres_warehouse: Engine,
    ):
        """Set up warehouse and database using fixtures."""
        # Store fixtures as class attributes for use in tests with self.*
        n_true_entities = 100

        self.__class__.client = matchbox_client
        self.__class__.warehouse_engine = postgres_warehouse
        self.__class__.n_true_entities = n_true_entities

        # Create feature configurations
        features = {
            "company_name": FeatureConfig(
                name="company_name",
                base_generator="company",
            ),
            "crn": FeatureConfig(
                name="crn",
                base_generator="bothify",
                parameters=(("text", "???-###-???-###"),),
            ),
            "duns": FeatureConfig(
                name="duns",
                base_generator="numerify",
                parameters=(("text", "########"),),
            ),
            "cdms": FeatureConfig(
                name="cdms",
                base_generator="numerify",
                parameters=(("text", "ORG-########"),),
            ),
        }

        # Create source configurations that match our test fixtures
        source_configs = (
            SourceConfig(
                full_name="e2e.crn",
                engine=postgres_warehouse,
                features=(
                    features["company_name"].add_variations(
                        SuffixRule(suffix=" Limited"),
                        SuffixRule(suffix=" UK"),
                        SuffixRule(suffix=" Company"),
                    ),
                    features["crn"],
                ),
                drop_base=True,
                n_true_entities=n_true_entities,
                repetition=0,  # No duplicates within the variations
            ),
            SourceConfig(
                full_name="e2e.duns",
                engine=postgres_warehouse,
                features=(
                    features["company_name"],
                    features["duns"],
                ),
                n_true_entities=n_true_entities // 2,  # Half the companies
                repetition=0,
            ),
            SourceConfig(
                full_name="e2e.cdms",
                engine=postgres_warehouse,
                features=(
                    features["crn"],
                    features["cdms"],
                ),
                n_true_entities=n_true_entities,
                repetition=1,  # Duplicate all rows
            ),
        )

        # Create linked sources testkit with our configurations
        self.__class__.linked_testkit = linked_sources_factory(
            source_configs=source_configs,
            seed=42,  # For reproducibility
        )

        # Use a separate schema to avoid conflict with legacy test data
        # TODO: Remove once legacy tests are refactored
        with postgres_warehouse.connect() as conn:
            conn.execute(text("CREATE SCHEMA IF NOT EXISTS e2e;"))
            conn.commit()

        # Setup code - Create tables in warehouse
        for source_testkit in self.linked_testkit.sources.values():
            source_testkit.to_warehouse(engine=postgres_warehouse)

        # Clear matchbox database before test
        response = matchbox_client.delete("/database", params={"certain": "true"})

        assert response.status_code == 200, "Failed to clear matchbox database"

        yield

        # Teardown code
        # Clean up database tables
        with postgres_warehouse.connect() as conn:
            for source_name in self.linked_testkit.sources:
                conn.execute(text(f"DROP TABLE IF EXISTS {source_name};"))
            conn.execute(text("DROP SCHEMA IF EXISTS e2e CASCADE;"))
            conn.commit()

        response = matchbox_client.delete("/database", params={"certain": "true"})
        assert response.status_code == 200, "Failed to clear matchbox database"

    def test_e2e_deduplication_and_linking_pipeline(self):
        """Runs an end to end test of the entire entity resolution pipeline.

        1. Create synthetic test data with linked_sources_factory
        2. Index the data in PostgreSQL
        3. Deduplicate each source, verify results
        4. Link pairs of deduplicated sources, verify results
        5. Link the linked pairs to create a full graph, verify results
        6. Verify the final linked data matches the original entity structure
        """
        # === SETUP PHASE ===
        # Get all sources and true entities for later verification
        all_true_entities = list(self.linked_testkit.true_entities)

        # Basic validation of our test data
        assert len(self.linked_testkit.sources) == 3, "Expected 3 sources"
        assert len(all_true_entities) == self.n_true_entities

        # Index all sources in the PostgreSQL database
        for source_testkit in self.linked_testkit.sources.values():
            source = source_testkit.source
            index(
                full_name=source.address.full_name,
                db_pk="pk",  # Primary key in our test data
                engine=self.warehouse_engine,
                columns=[col.model_dump() for col in source.columns],
            )
            logging.debug(f"Indexed source: {source.address.full_name}")

        # Helper functions
        # Define custom company name cleaner, mirroring the FeatureConfig
        remove_stopwords_redux = partial(
            steps.remove_stopwords, stopwords=["Limited", "UK", "Company"]
        )
        clean_company_name = cleaning_function(
            steps.tokenise,  # returns array
            remove_stopwords_redux,
            steps.list_join_to_string,  # returns col
            steps.trim,
        )

        def _clean_company_name(df: DataFrame, prefix: str) -> DataFrame:
            """Clean company_name feature in the source or model."""
            if any("company_name" in col for col in df.columns):
                company_cleaner = cleaner(
                    function=clean_company_name,
                    arguments={"column": f"{prefix}company_name"},
                )
                clean_pipeline = cleaners(company_cleaner)
                return process(data=df, pipeline=clean_pipeline)

            return df

        # === DEDUPLICATION PHASE ===
        deduper_names = {}

        for source_name, source_testkit in self.linked_testkit.sources.items():
            # Get prefix for column names
            prefix = fullname_to_prefix(source_name)

            # Query data from the source
            # PK included then dropped to create ClusterEntity objects for later diff
            source_select = select(
                {
                    source_name: ["pk"]
                    + [col.name for col in source_testkit.source.columns]
                },
                engine=self.warehouse_engine,
            )
            raw_df = query(source_select, return_type="pandas")
            clusters = query_to_cluster_entities(
                query=raw_df,
                source_pks={source_name: f"{prefix}pk"},
            )
            df = raw_df.drop(f"{prefix}pk", axis=1)

            # Apply cleaning based on features in the source
            cleaned = _clean_company_name(df, prefix)

            # Get feature names with prefix for deduplication
            feature_names = [
                f"{prefix}{feature.name}" for feature in source_testkit.features
            ]

            # Create and run a deduper model
            deduper_name = f"deduper_{source_name}"
            deduper = make_model(
                model_name=deduper_name,
                description=f"Deduplication of {source_name}",
                model_class=NaiveDeduper,
                model_settings={
                    "id": "id",
                    "unique_fields": feature_names,
                },
                left_data=cleaned,
                left_resolution=source_testkit.source.resolution_name,
            )

            # Run the deduper and store results
            results = deduper.run()

            # Verify deduplication results match expected true entities
            identical, report = self.linked_testkit.diff_results(
                probabilities=results.probabilities,
                left_clusters=clusters,
                right_clusters=None,
                sources=[source_name],
                threshold=0,
            )

            assert identical, f"Deduplication of {source_name} failed: {report}"

            results.to_matchbox()
            deduper.truth = 1.0

            logging.debug(f"Successfully deduplicated {source_name}")

            # Store the deduper resolution name for later use
            deduper_names[source_name] = deduper_name

        # === PAIRWISE LINKING PHASE ===
        linker_names = {}

        # Define linking pairs based on common fields
        linking_pairs = [
            (
                self.linked_testkit.sources["e2e.crn"],
                self.linked_testkit.sources["e2e.duns"],
                "company_name",
            ),  # CRN-DUNS via company_name
            (
                self.linked_testkit.sources["e2e.crn"],
                self.linked_testkit.sources["e2e.cdms"],
                "crn",
            ),  # CRN-CDMS via crn
        ]

        for left_testkit, right_testkit, common_field in linking_pairs:
            # Get prefixes for column names
            left_prefix = fullname_to_prefix(left_testkit.source.address.full_name)
            right_prefix = fullname_to_prefix(right_testkit.source.address.full_name)

            # Query deduplicated data
            # PK included then dropped to create ClusterEntity objects for later diff
            left_raw_df = query(
                select(
                    {left_testkit.source.address.full_name: ["pk", common_field]},
                    engine=self.warehouse_engine,
                ),
                resolution_name=deduper_names[left_testkit.source.address.full_name],
                return_type="pandas",
            )
            left_clusters = query_to_cluster_entities(
                query=left_raw_df,
                source_pks={left_testkit.source.address.full_name: f"{left_prefix}pk"},
            )
            left_df = left_raw_df.drop(f"{left_prefix}pk", axis=1)

            right_raw_df = query(
                select(
                    {right_testkit.source.address.full_name: ["pk", common_field]},
                    engine=self.warehouse_engine,
                ),
                resolution_name=deduper_names[right_testkit.source.address.full_name],
                return_type="pandas",
            )
            right_clusters = query_to_cluster_entities(
                query=right_raw_df,
                source_pks={
                    right_testkit.source.address.full_name: f"{right_prefix}pk"
                },
            )
            right_df = right_raw_df.drop(f"{right_prefix}pk", axis=1)

            # Apply cleaning based on features in the sources
            left_cleaned = _clean_company_name(left_df, left_prefix)
            right_cleaned = _clean_company_name(right_df, right_prefix)

            # Build comparison clause
            comparison_clause = (
                f"l.{left_prefix}{common_field} = r.{right_prefix}{common_field}"
            )

            # Create and run linker model
            linker_name = (
                f"linker_{left_testkit.source.address.full_name}"
                f"_{right_testkit.source.address.full_name}"
            )
            linker = make_model(
                model_name=linker_name,
                description=f"Linking {left_testkit.name} and {right_testkit.name}",
                model_class=DeterministicLinker,
                model_settings={
                    "left_id": "id",
                    "right_id": "id",
                    "comparisons": comparison_clause,
                },
                left_data=left_cleaned,
                left_resolution=deduper_names[left_testkit.source.address.full_name],
                right_data=right_cleaned,
                right_resolution=deduper_names[right_testkit.source.address.full_name],
            )

            # Run the linker and store results
            results = linker.run()

            # Verify linking results
            identical, report = self.linked_testkit.diff_results(
                probabilities=results.probabilities,
                left_clusters=left_clusters,
                right_clusters=right_clusters,
                sources=[
                    left_testkit.source.address.full_name,
                    right_testkit.source.address.full_name,
                ],
                threshold=0,
            )

            report_counts = {
                k: len(v) if isinstance(v, list) else v for k, v in report.items()
            }

            assert identical, (
                f"Linking of {left_testkit.name} to {right_testkit.name} failed: "
                f"{report_counts}"
            )

            results.to_matchbox()
            linker.truth = 1.0

            logging.debug(
                f"Successfully linked {left_testkit.name} and {right_testkit.name}"
            )

            # Store the linker resolution name for later use
            linker_names[
                (
                    left_testkit.source.address.full_name,
                    right_testkit.source.address.full_name,
                )
            ] = linker_name

        # === FINAL LINKING PHASE ===
        # Now link the first linked pair (crn-duns) with the third source (cdms)
        crn_source = "e2e.crn"
        duns_source = "e2e.duns"
        cdms_source = "e2e.cdms"
        first_pair = (crn_source, duns_source)

        # Get prefixes for column names
        crn_prefix = fullname_to_prefix(crn_source)
        duns_prefix = fullname_to_prefix(duns_source)
        cdms_prefix = fullname_to_prefix(cdms_source)

        # Query data from the first linked pair and the third source
        # PK included then dropped to create ClusterEntity objects for later diff
        left_raw_df = query(
            select({crn_source: ["pk", "crn"]}, engine=self.warehouse_engine),
            select({duns_source: ["pk"]}, engine=self.warehouse_engine),
            resolution_name=linker_names[first_pair],
            return_type="pandas",
        )
        left_clusters = query_to_cluster_entities(
            query=left_raw_df,
            source_pks={crn_source: f"{crn_prefix}pk", duns_source: f"{duns_prefix}pk"},
        )
        left_df = left_raw_df.drop(f"{left_prefix}pk", axis=1)

        right_raw_df = query(
            select({cdms_source: ["pk", "crn"]}, engine=self.warehouse_engine),
            resolution_name=deduper_names[cdms_source],
            return_type="pandas",
        )
        right_clusters = query_to_cluster_entities(
            query=right_raw_df, source_pks={cdms_source: f"{cdms_prefix}pk"}
        )
        right_df = right_raw_df.drop(f"{right_prefix}pk", axis=1)

        # Apply cleaning if needed
        left_cleaned = _clean_company_name(left_df, crn_prefix)
        right_cleaned = _clean_company_name(right_df, cdms_prefix)

        # Create and run final linker with the common "crn" field
        final_linker_name = "final_linker"
        final_linker = make_model(
            model_name=final_linker_name,
            description="Final linking of all sources",
            model_class=DeterministicLinker,
            model_settings={
                "left_id": "id",
                "right_id": "id",
                "comparisons": f"l.{crn_prefix}crn = r.{cdms_prefix}crn",
            },
            left_data=left_cleaned,
            left_resolution=linker_names[first_pair],
            right_data=right_cleaned,
            right_resolution=deduper_names[cdms_source],
        )

        # Run the final linker and store results
        results = final_linker.run()

        # Verify final linking results
        identical, report = self.linked_testkit.diff_results(
            probabilities=results.probabilities,
            left_clusters=left_clusters,
            right_clusters=right_clusters,
            sources=[crn_source, duns_source, cdms_source],
            threshold=0,
        )

        report_counts = {
            k: len(v) if isinstance(v, list) else v for k, v in report.items()
        }

        assert identical, f"Final linking failed: {report_counts}"

        results.to_matchbox()
        final_linker.truth = 1.0

        logging.debug("Successfully linked all sources")

        # === FINAL VERIFICATION PHASE ===
        # Query the final linked data with specific columns
        crn_source = "e2e.crn"
        duns_source = "e2e.duns"
        cdms_source = "e2e.cdms"

        # Get necessary columns from each source
        final_df = query(
            select(
                {
                    crn_source: ["pk", "company_name", "crn"],
                    duns_source: ["pk", "company_name", "duns"],
                    cdms_source: ["pk", "crn", "cdms"],
                },
                engine=self.warehouse_engine,
            ),
            resolution_name=final_linker_name,
            return_type="pandas",
        )

        final_clusters = query_to_cluster_entities(
            query=final_df,
            source_pks={
                crn_source: f"{crn_prefix}pk",
                duns_source: f"{duns_prefix}pk",
                cdms_source: f"{cdms_prefix}pk",
            },
        )

        # Verify the final data structure - number of unique entities
        assert (
            final_df["id"].nunique()
            == len(self.linked_testkit.true_entities)
            == self.n_true_entities
        ), (
            f"Expected {len(self.linked_testkit.true_entities)} unique entities, "
            f"got {final_df['id'].nunique()}"
        )

        # Verify the final cluster membership -- the golden check

        true_entities = {
            entity.to_cluster_entity(crn_source, duns_source, cdms_source)
            for entity in self.linked_testkit.true_entities
            if entity.to_cluster_entity(crn_source, duns_source, cdms_source)
            is not None
        }

        assert true_entities == set(final_clusters), "Final clusters do not match"

        logging.debug("E2E test completed successfully!")



from functools import partial

import pyarrow as pa
import pyarrow.compute as pc
import pytest
from sqlalchemy import Engine

from matchbox.common.dtos import ModelAncestor, ModelMetadata, ModelType
from matchbox.common.exceptions import (
    MatchboxDataNotFound,
    MatchboxResolutionNotFoundError,
    MatchboxSourceNotFoundError,
)
from matchbox.common.factories.entities import SourceEntity
from matchbox.common.factories.sources import SourceTestkit
from matchbox.common.graph import ResolutionGraph
from matchbox.common.hash import HASH_FUNC
from matchbox.common.sources import Match, SourceAddress
from matchbox.server.base import MatchboxDBAdapter

from ..fixtures.db import setup_scenario

backends = [
    pytest.param("matchbox_postgres", id="postgres"),
]


@pytest.fixture(scope="function")
def backend_instance(request: pytest.FixtureRequest, backend: str):
    """Create a fresh backend instance for each test."""
    backend_obj = request.getfixturevalue(backend)
    backend_obj.clear(certain=True)
    return backend_obj


@pytest.mark.parametrize("backend", backends)
@pytest.mark.docker
class TestMatchboxBackend:
    @pytest.fixture(autouse=True)
    def setup(self, backend_instance: str, sqlite_warehouse: Engine):
        self.backend: MatchboxDBAdapter = backend_instance
        self.scenario = partial(setup_scenario, warehouse=sqlite_warehouse)

    def test_properties(self):
        """Test that properties obey their protocol restrictions."""
        with self.scenario(self.backend, "index"):
            assert isinstance(self.backend.datasets.list_all(), list)
            assert isinstance(self.backend.datasets.count(), int)
            assert isinstance(self.backend.models.count(), int)
            assert isinstance(self.backend.data.count(), int)
            assert isinstance(self.backend.clusters.count(), int)
            assert isinstance(self.backend.creates.count(), int)
            assert isinstance(self.backend.merges.count(), int)
            assert isinstance(self.backend.proposes.count(), int)

    def test_validate_ids(self):
        """Test validating data IDs."""
        with self.scenario(self.backend, "dedupe") as dag:
            crn_testkit = dag.sources.get("crn")

            df_crn = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name="naive_test.crn",
            )

            ids = df_crn["id"].to_pylist()
            assert len(ids) > 0
            self.backend.validate_ids(ids=ids)

            with pytest.raises(MatchboxDataNotFound):
                self.backend.validate_ids(ids=[-6])

    def test_validate_hashes(self):
        """Test validating data hashes."""
        with self.scenario(self.backend, "dedupe") as dag:
            crn_testkit = dag.sources.get("crn")

            df_crn = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name="naive_test.crn",
            )

            ids = df_crn["id"].to_pylist()
            hashes = list(self.backend.cluster_id_to_hash(ids=ids).values())
            assert len(hashes) > 0
            self.backend.validate_hashes(hashes=hashes)

            with pytest.raises(MatchboxDataNotFound):
                self.backend.validate_hashes(
                    hashes=[HASH_FUNC(b"nonexistent").digest()]
                )

    def test_cluster_id_to_hash(self):
        """Test getting ID to Cluster hash lookup from the database."""
        with self.scenario(self.backend, "dedupe") as dag:
            crn_testkit = dag.sources.get("crn")

            df_crn = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name="naive_test.crn",
            )

            ids = df_crn["id"].to_pylist()
            assert len(ids) > 0

            hashes = self.backend.cluster_id_to_hash(ids=ids)
            assert len(hashes) == len(set(ids))
            assert set(ids) == set(hashes.keys())
            assert all(isinstance(h, bytes) for h in hashes.values())

            assert self.backend.cluster_id_to_hash(ids=[-6]) == {-6: None}

    def test_get_source(self):
        """Test querying data from the database."""
        with self.scenario(self.backend, "index") as dag:
            crn_testkit = dag.sources.get("crn")

            crn_retrieved = self.backend.get_source(crn_testkit.source.address)
            # Equality between the two is False because one lacks the Engine
            assert crn_testkit.source.model_dump() == crn_retrieved.model_dump()

            with pytest.raises(MatchboxSourceNotFoundError):
                self.backend.get_source(
                    SourceAddress(
                        full_name="foo", warehouse_hash=bytes("bar".encode("ascii"))
                    )
                )

    def test_get_resolution_sources(self):
        """Test retrieving sources available to a resolution."""
        with self.scenario(self.backend, "link") as dag:
            crn, duns = dag.sources["crn"], dag.sources["duns"]
            dedupe_sources = self.backend.get_resolution_sources(
                resolution_name="naive_test.crn"
            )
            link_sources = self.backend.get_resolution_sources(
                resolution_name="deterministic_naive_test.crn_naive_test.duns"
            )

            assert {s.resolution_name for s in dedupe_sources} == {
                crn.source.resolution_name
            }

            assert {s.resolution_name for s in link_sources} == {
                crn.source.resolution_name,
                duns.source.resolution_name,
            }

    def test_get_resolution_graph(self):
        """Test getting the resolution graph."""
        graph = self.backend.get_resolution_graph()
        assert len(graph.nodes) == 0
        assert len(graph.edges) == 0
        assert isinstance(graph, ResolutionGraph)

        with self.scenario(self.backend, "link"):
            graph = self.backend.get_resolution_graph()
            # Nodes: 3 datasets, 3 dedupers, and 3 linkers
            # Edges: 1 per deduper, 2 per linker
            assert len(graph.nodes) == 9
            assert len(graph.edges) == 9

    def test_get_model(self):
        """Test getting a model from the database."""
        with self.scenario(self.backend, "dedupe"):
            model = self.backend.get_model(model="naive_test.crn")
            assert isinstance(model, ModelMetadata)

            with pytest.raises(MatchboxResolutionNotFoundError):
                self.backend.get_model(model="nonexistent")

    def test_delete_model(self):
        """
        Tests the deletion of:

        * The model from the model table
        * The creates edges the model made
        * Any models that depended on this model, and their creates edges
        * Any probability values associated with the model
        * All of the above for all parent models. As every model is defined by
            its parents, deleting a model means cascading deletion to all descendants
        """
        with self.scenario(self.backend, "link") as dag:
            # Expect it to delete itself, its probabilities,
            # its parents, and their probabilities
            deduper_to_delete = "naive_test.crn"
            total_models = len(dag.models)

            models_pre_delete = self.backend.models.count()
            cluster_count_pre_delete = self.backend.clusters.count()
            cluster_assoc_count_pre_delete = self.backend.creates.count()
            proposed_merge_probs_pre_delete = self.backend.proposes.count()
            actual_merges_pre_delete = self.backend.merges.count()

            assert models_pre_delete == total_models
            assert cluster_count_pre_delete > 0
            assert cluster_assoc_count_pre_delete > 0
            assert proposed_merge_probs_pre_delete > 0
            assert actual_merges_pre_delete > 0

            # Perform deletion
            self.backend.delete_model(deduper_to_delete, certain=True)

            models_post_delete = self.backend.models.count()
            cluster_count_post_delete = self.backend.clusters.count()
            cluster_assoc_count_post_delete = self.backend.creates.count()
            proposed_merge_probs_post_delete = self.backend.proposes.count()
            actual_merges_post_delete = self.backend.merges.count()

            # Deletes deduper and parent linkers: 4 models gone
            assert models_post_delete == models_pre_delete - 4

            # Cluster, dedupe and link count unaffected
            assert cluster_count_post_delete == cluster_count_pre_delete
            assert actual_merges_post_delete == actual_merges_pre_delete

            # But count of propose and create edges has dropped
            assert cluster_assoc_count_post_delete < cluster_assoc_count_pre_delete
            assert proposed_merge_probs_post_delete < proposed_merge_probs_pre_delete

    def test_insert_model(self):
        """Test that models can be inserted."""
        with self.scenario(self.backend, "index") as dag:
            crn_testkit = dag.sources.get("crn")
            duns_testkit = dag.sources.get("duns")

            # Test deduper insertion
            models_count = self.backend.models.count()

            self.backend.insert_model(
                model=ModelMetadata(
                    name="dedupe_1",
                    description="Test deduper 1",
                    type=ModelType.DEDUPER,
                    left_resolution=crn_testkit.source.resolution_name,
                )
            )
            self.backend.insert_model(
                model=ModelMetadata(
                    name="dedupe_2",
                    description="Test deduper 2",
                    type=ModelType.DEDUPER,
                    left_resolution=duns_testkit.source.resolution_name,
                )
            )

            assert self.backend.models.count() == models_count + 2

            # Test linker insertion
            self.backend.insert_model(
                model=ModelMetadata(
                    name="link_1",
                    description="Test linker 1",
                    type=ModelType.LINKER,
                    left_resolution="dedupe_1",
                    right_resolution="dedupe_2",
                )
            )

            assert self.backend.models.count() == models_count + 3

            # Test model upsert
            self.backend.insert_model(
                model=ModelMetadata(
                    name="link_1",
                    description="Test upsert",
                    type=ModelType.LINKER,
                    left_resolution="dedupe_1",
                    right_resolution="dedupe_2",
                )
            )

            assert self.backend.models.count() == models_count + 3

    def test_model_results_basic(self):
        """Test that a model's results data can be set and retrieved."""
        with self.scenario(self.backend, "dedupe"):
            # Retrieve
            pre_results = self.backend.get_model_results(model="naive_test.crn")

            assert isinstance(pre_results, pa.Table)
            assert len(pre_results) > 0

            self.backend.validate_ids(ids=pre_results["id"].to_pylist())
            self.backend.validate_ids(ids=pre_results["left_id"].to_pylist())
            self.backend.validate_ids(ids=pre_results["right_id"].to_pylist())

            # Set
            target_row = pre_results.to_pylist()[0]
            target_id = target_row["id"]
            target_left_id = target_row["left_id"]
            target_right_id = target_row["right_id"]

            matches_id_mask = pc.not_equal(pre_results["id"], target_id)
            matches_left_mask = pc.not_equal(pre_results["left_id"], target_left_id)
            matches_right_mask = pc.not_equal(pre_results["right_id"], target_right_id)

            combined_mask = pc.and_(
                pc.and_(matches_id_mask, matches_left_mask), matches_right_mask
            )
            df_probabilities_truncated = pre_results.filter(combined_mask)

            results = df_probabilities_truncated.select(
                ["left_id", "right_id", "probability"]
            )

            self.backend.set_model_results(model="naive_test.crn", results=results)

            # Retrieve again
            post_results = self.backend.get_model_results(model="naive_test.crn")

            # Check difference
            assert len(pre_results) != len(post_results)

    def test_model_results_shared_clusters(self):
        """Test that model results data can be inserted when clusters are shared."""
        with self.scenario(self.backend, "convergent") as dag:
            for model_testkit in dag.models.values():
                self.backend.insert_model(model=model_testkit.model.metadata)
                self.backend.set_model_results(
                    model=model_testkit.name, results=model_testkit.probabilities
                )

    def test_model_truth(self):
        """Test that a model's truth can be set and retrieved."""
        with self.scenario(self.backend, "dedupe"):
            # Retrieve
            pre_truth = self.backend.get_model_truth(model="naive_test.crn")

            # Set
            self.backend.set_model_truth(model="naive_test.crn", truth=75)

            # Retrieve again
            post_truth = self.backend.get_model_truth(model="naive_test.crn")

            # Check difference
            assert pre_truth != post_truth

    def test_model_ancestors(self):
        """Test that a model's ancestors can be retrieved."""
        with self.scenario(self.backend, "link"):
            linker_name = "deterministic_naive_test.crn_naive_test.duns"
            linker_ancestors = self.backend.get_model_ancestors(model=linker_name)

            assert [
                isinstance(ancestor, ModelAncestor) for ancestor in linker_ancestors
            ]

            # Not all ancestors have truth values, but one must
            truth_found = False
            for ancestor in linker_ancestors:
                if isinstance(ancestor.truth, int):
                    truth_found = True

            assert truth_found

    def test_model_ancestors_cache(self):
        """Test that a model's ancestors cache can be set and retrieved."""
        with self.scenario(self.backend, "link"):
            linker_name = "deterministic_naive_test.crn_naive_test.duns"

            # Retrieve
            pre_ancestors_cache = self.backend.get_model_ancestors_cache(
                model=linker_name
            )

            # Set
            updated_ancestors_cache = [
                ModelAncestor(name=ancestor.name, truth=90)
                for ancestor in pre_ancestors_cache
            ]
            self.backend.set_model_ancestors_cache(
                model=linker_name, ancestors_cache=updated_ancestors_cache
            )

            # Retrieve again
            post_ancestors_cache = self.backend.get_model_ancestors_cache(
                model=linker_name
            )

            # Check difference
            assert pre_ancestors_cache != post_ancestors_cache
            assert post_ancestors_cache == updated_ancestors_cache

    def test_index(self):
        """Test that indexing data works."""
        assert self.backend.data.count() == 0

        with self.scenario(self.backend, "index") as dag:
            assert self.backend.data.count() == (
                len(dag.sources["crn"].entities)
                + len(dag.sources["cdms"].entities)
                + len(dag.sources["duns"].entities)
            )

    def test_index_new_source(self):
        """Test that indexing identical works."""
        with self.scenario(self.backend, "bare") as dag:
            crn_testkit: SourceTestkit = dag.sources.get("crn")

            assert self.backend.clusters.count() == 0

            self.backend.index(crn_testkit.source, crn_testkit.data_hashes)

            crn_retrieved = self.backend.get_source(crn_testkit.source.address)

            # Equality between the two is False because one lacks the Engine
            assert crn_testkit.source.model_dump() == crn_retrieved.model_dump()
            assert self.backend.data.count() == len(crn_testkit.data_hashes)
            # I can add it again with no consequences
            self.backend.index(crn_testkit.source, crn_testkit.data_hashes)
            assert self.backend.data.count() == len(crn_testkit.data_hashes)
            assert self.backend.source_resolutions.count() == 1

    def test_index_duplicate_clusters(self):
        """Test that indexing new data with duplicate hashes works."""
        with self.scenario(self.backend, "bare") as dag:
            crn_testkit: SourceTestkit = dag.sources.get("crn")

            data_hashes_halved = crn_testkit.data_hashes.slice(
                0, crn_testkit.data_hashes.num_rows // 2
            )

            assert self.backend.data.count() == 0
            self.backend.index(crn_testkit.source, data_hashes_halved)
            assert self.backend.data.count() == data_hashes_halved.num_rows
            self.backend.index(crn_testkit.source, crn_testkit.data_hashes)
            assert self.backend.data.count() == crn_testkit.data_hashes.num_rows
            assert self.backend.source_resolutions.count() == 1

    def test_index_same_resolution(self):
        """Test that indexing same-name sources in different warehouses works."""
        with self.scenario(self.backend, "bare") as dag:
            crn_testkit: SourceTestkit = dag.sources.get("crn")

            crn_source_1 = crn_testkit.source.model_copy(
                update={
                    "address": SourceAddress(
                        full_name=crn_testkit.source.address.full_name,
                        warehouse_hash=b"bar1",
                    )
                }
            )
            crn_source_2 = crn_testkit.source.model_copy(
                deep=True,
                update={
                    "address": SourceAddress(
                        full_name=crn_testkit.source.address.full_name,
                        warehouse_hash=b"bar2",
                    )
                },
            )

            self.backend.index(crn_source_1, crn_testkit.data_hashes)
            self.backend.index(crn_source_2, crn_testkit.data_hashes)

            assert self.backend.data.count() == len(crn_testkit.data_hashes)
            assert self.backend.source_resolutions.count() == 1

    def test_index_different_resolution_same_hashes(self):
        """Test that indexing data with the same hashes but different sources works."""
        with self.scenario(self.backend, "bare") as dag:
            crn_testkit: SourceTestkit = dag.sources.get("crn")
            duns_testkit: SourceTestkit = dag.sources.get("duns")

            self.backend.index(crn_testkit.source, crn_testkit.data_hashes)
            # Different source, same data
            self.backend.index(duns_testkit.source, crn_testkit.data_hashes)
            assert self.backend.data.count() == len(crn_testkit.data_hashes)
            assert self.backend.source_resolutions.count() == 2

    def test_query_only_source(self):
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "index") as dag:
            crn_testkit = dag.sources.get("crn")

            df_crn_sample = self.backend.query(
                source_address=crn_testkit.source.address,
                limit=10,
            )

            assert isinstance(df_crn_sample, pa.Table)
            assert df_crn_sample.num_rows == 10

            df_crn_full = self.backend.query(source_address=crn_testkit.source.address)

            assert df_crn_full.num_rows == crn_testkit.query.num_rows
            assert set(df_crn_full.column_names) == {"id", "source_pk"}

    def test_query_with_dedupe_model(self):
        """Test querying data from a deduplication point of truth."""
        with self.scenario(self.backend, "dedupe") as dag:
            crn_testkit = dag.sources.get("crn")

            df_crn = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name="naive_test.crn",
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.query.num_rows
            assert set(df_crn.column_names) == {"id", "source_pk"}

            sources_dict = dag.get_sources_for_model("naive_test.crn")
            assert len(sources_dict) == 1
            linked = dag.linked[next(iter(sources_dict))]

            assert pc.count_distinct(df_crn["id"]).as_py() == len(
                linked.true_entity_subset("crn")
            )

    def test_query_with_link_model(self):
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "link") as dag:
            linker_name = "deterministic_naive_test.crn_naive_test.duns"
            crn_testkit = dag.sources.get("crn")
            duns_testkit = dag.sources.get("duns")

            df_crn = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name=linker_name,
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.query.num_rows
            assert set(df_crn.column_names) == {"id", "source_pk"}

            df_duns = self.backend.query(
                source_address=duns_testkit.source.address,
                resolution_name=linker_name,
            )

            assert isinstance(df_duns, pa.Table)
            assert df_duns.num_rows == duns_testkit.query.num_rows
            assert set(df_duns.column_names) == {"id", "source_pk"}

            sources_dict = dag.get_sources_for_model(linker_name)
            assert len(sources_dict) == 1
            linked = dag.linked[next(iter(sources_dict))]

            all_ids = pa.concat_arrays(
                [df_crn["id"].combine_chunks(), df_duns["id"].combine_chunks()]
            )

            assert pc.count_distinct(all_ids).as_py() == len(
                linked.true_entity_subset("crn", "duns")
            )

    def test_threshold_query_with_link_model(self):
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "link") as dag:
            linker_name = "probabilistic_naive_test.crn_naive_test.cdms"
            crn_testkit = dag.sources.get("crn")
            cdms_testkit = dag.sources.get("cdms")

            df_crn = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name=linker_name,
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.query.num_rows
            assert set(df_crn.column_names) == {"id", "source_pk"}

            df_cdms = self.backend.query(
                source_address=cdms_testkit.source.address,
                resolution_name=linker_name,
            )

            assert isinstance(df_cdms, pa.Table)
            assert df_cdms.num_rows == cdms_testkit.query.num_rows
            assert set(df_cdms.column_names) == {"id", "source_pk"}

            sources_dict = dag.get_sources_for_model(linker_name)
            assert len(sources_dict) == 1
            linked = dag.linked[next(iter(sources_dict))]

            # Test query with threshold
            df_crn_threshold = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name=linker_name,
                threshold=100,
            )
            df_cdms_threshold = self.backend.query(
                source_address=cdms_testkit.source.address,
                resolution_name=linker_name,
                threshold=100,
            )
            threshold_ids = pa.concat_arrays(
                [
                    df_crn_threshold["id"].combine_chunks(),
                    df_cdms_threshold["id"].combine_chunks(),
                ]
            )

            # Query returns more clusters when threshold exceeds
            # true entity match probabilities
            assert pc.count_distinct(threshold_ids).as_py() > len(
                linked.true_entity_subset("crn", "cdms")
            )

    def test_match_one_to_many(self):
        """Test that matching data works when the target has many IDs."""
        with self.scenario(self.backend, "link") as dag:
            linker_name = "deterministic_naive_test.crn_naive_test.duns"
            crn_testkit = dag.sources.get("crn")
            duns_testkit = dag.sources.get("duns")

            sources_dict = dag.get_sources_for_model(linker_name)
            assert len(sources_dict) == 1
            linked = dag.linked[next(iter(sources_dict))]

            # A random one:many entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "duns": 1},
                max_appearances={"duns": 1},
            )[0]

            res = self.backend.match(
                source_pk=next(iter(source_entity.source_pks["duns"])),
                source=duns_testkit.source.address,
                targets=[crn_testkit.source.address],
                resolution_name=linker_name,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == duns_testkit.source.address
            assert res[0].target == crn_testkit.source.address
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.source_pks["duns"]
            assert res[0].target_id == source_entity.source_pks["crn"]

    def test_match_many_to_one(self):
        """Test that matching data works when the source has more possible IDs."""
        with self.scenario(self.backend, "link") as dag:
            linker_name = "deterministic_naive_test.crn_naive_test.duns"
            crn_testkit = dag.sources.get("crn")
            duns_testkit = dag.sources.get("duns")

            sources_dict = dag.get_sources_for_model(linker_name)
            assert len(sources_dict) == 1
            linked = dag.linked[next(iter(sources_dict))]

            # A random many:one entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "duns": 1},
                max_appearances={"duns": 1},
            )[0]

            res = self.backend.match(
                source_pk=next(iter(source_entity.source_pks["crn"])),
                source=crn_testkit.source.address,
                targets=[duns_testkit.source.address],
                resolution_name=linker_name,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.address
            assert res[0].target == duns_testkit.source.address
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.source_pks["crn"]
            assert res[0].target_id == source_entity.source_pks["duns"]

    def test_match_one_to_none(self):
        """Test that matching data works when the target has no IDs."""
        with self.scenario(self.backend, "link") as dag:
            linker_name = "deterministic_naive_test.crn_naive_test.duns"
            crn_testkit = dag.sources.get("crn")
            duns_testkit = dag.sources.get("duns")

            sources_dict = dag.get_sources_for_model(linker_name)
            assert len(sources_dict) == 1
            linked = dag.linked[next(iter(sources_dict))]

            # A random one:none entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 1},
                max_appearances={"duns": 0},
            )[0]

            res = self.backend.match(
                source_pk=next(iter(source_entity.source_pks["crn"])),
                source=crn_testkit.source.address,
                targets=[duns_testkit.source.address],
                resolution_name=linker_name,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.address
            assert res[0].target == duns_testkit.source.address
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.source_pks["crn"]
            assert res[0].target_id == source_entity.source_pks.get("duns", set())

    def test_match_none_to_none(self):
        """Test that matching data works when the supplied key doesn't exist."""
        with self.scenario(self.backend, "link") as dag:
            linker_name = "deterministic_naive_test.crn_naive_test.duns"
            crn_testkit = dag.sources.get("crn")
            duns_testkit = dag.sources.get("duns")

            # Use a non-existent source primary key
            non_existent_pk = "foo"

            res = self.backend.match(
                source_pk=non_existent_pk,
                source=crn_testkit.source.address,
                targets=[duns_testkit.source.address],
                resolution_name=linker_name,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.address
            assert res[0].target == duns_testkit.source.address
            assert res[0].cluster is None
            assert res[0].source_id == set()
            assert res[0].target_id == set()

    def test_threshold_match_many_to_one(self):
        """Test that matching data works when the target has many IDs."""
        with self.scenario(self.backend, "link") as dag:
            linker_name = "probabilistic_naive_test.crn_naive_test.cdms"
            crn_testkit = dag.sources.get("crn")
            cdms_testkit = dag.sources.get("cdms")

            sources_dict = dag.get_sources_for_model(linker_name)
            assert len(sources_dict) == 1
            linked = dag.linked[next(iter(sources_dict))]

            # A random one:many entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "cdms": 1},
                max_appearances={"cdms": 1},
            )[0]

            res = self.backend.match(
                source_pk=next(iter(source_entity.source_pks["crn"])),
                source=crn_testkit.source.address,
                targets=[cdms_testkit.source.address],
                resolution_name=linker_name,
                threshold=100,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.address
            assert res[0].target == cdms_testkit.source.address
            assert res[0].source_id == source_entity.source_pks["crn"]
            # Match does not return true target ids when threshold
            # exceeds match probability
            assert len(res[0].target_id) < len(source_entity.source_pks["cdms"])

    def test_clear(self):
        """Test deleting all rows in the database."""
        with self.scenario(self.backend, "dedupe"):
            assert self.backend.datasets.count() > 0
            assert self.backend.data.count() > 0
            assert self.backend.models.count() > 0
            assert self.backend.clusters.count() > 0
            assert self.backend.creates.count() > 0
            assert self.backend.merges.count() > 0
            assert self.backend.proposes.count() > 0

            self.backend.clear(certain=True)

            assert self.backend.datasets.count() == 0
            assert self.backend.data.count() == 0
            assert self.backend.models.count() == 0
            assert self.backend.clusters.count() == 0
            assert self.backend.creates.count() == 0
            assert self.backend.merges.count() == 0
            assert self.backend.proposes.count() == 0

    def test_clear_and_restore(self):
        """Test that clearing and restoring the database works."""
        with self.scenario(self.backend, "link") as dag:
            crn_testkit = dag.sources.get("crn")

            # Verify we have data
            pre_dump_datasets_count = self.backend.datasets.count()
            pre_dump_models_count = self.backend.models.count()
            pre_dump_data_count = self.backend.data.count()
            pre_dump_clusters_count = self.backend.clusters.count()
            pre_dump_merges_count = self.backend.merges.count()
            pre_dump_creates_count = self.backend.creates.count()
            pre_dump_proposes_count = self.backend.proposes.count()

            # All these should be greater than zero after setup
            assert pre_dump_datasets_count > 0
            assert pre_dump_models_count > 0
            assert pre_dump_data_count > 0
            assert pre_dump_clusters_count > 0
            assert pre_dump_merges_count > 0
            assert pre_dump_creates_count > 0
            assert pre_dump_proposes_count > 0

            # Get some specific IDs to verify they're restored properly
            df_crn_before = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name="naive_test.crn",
            )
            sample_ids_before = df_crn_before["id"].to_pylist()[:5]  # Take first 5 IDs

            # Dump the database
            snapshot = self.backend.dump()

            # Clear the database
            self.backend.clear(certain=True)

            # Verify database is empty
            assert self.backend.datasets.count() == 0
            assert self.backend.models.count() == 0
            assert self.backend.data.count() == 0
            assert self.backend.clusters.count() == 0
            assert self.backend.merges.count() == 0
            assert self.backend.creates.count() == 0
            assert self.backend.proposes.count() == 0

            # Restore from snapshot
            self.backend.restore(snapshot)

            # Verify counts match pre-dump state
            assert self.backend.datasets.count() == pre_dump_datasets_count
            assert self.backend.models.count() == pre_dump_models_count
            assert self.backend.data.count() == pre_dump_data_count
            assert self.backend.clusters.count() == pre_dump_clusters_count
            assert self.backend.merges.count() == pre_dump_merges_count
            assert self.backend.creates.count() == pre_dump_creates_count
            assert self.backend.proposes.count() == pre_dump_proposes_count

            # Verify specific data was restored correctly
            df_crn_after = self.backend.query(
                source_address=crn_testkit.source.address,
                resolution_name="naive_test.crn",
            )
            sample_ids_after = df_crn_after["id"].to_pylist()[:5]  # Take first 5 IDs

            # The same IDs should be present after restoration
            assert set(sample_ids_before) == set(sample_ids_after)

            # Test the clear parameter of restore
            self.backend.restore(snapshot, clear=True)

            # Verify counts still match
            assert self.backend.datasets.count() == pre_dump_datasets_count
            assert self.backend.models.count() == pre_dump_models_count
            assert self.backend.data.count() == pre_dump_data_count



"""Functions to pre-process data sources."""

from typing import Any, Callable

from pandas import DataFrame


def cleaner(function: Callable, arguments: dict) -> dict[str, dict[str, Any]]:
    """Define a function to clean a dataset.

    Args:
        function: the callable implementing the cleaning behaviour
        arguments: a dictionary of keyword arguments to pass to the cleaning function

    Returns:
        A representation of the cleaner ready to be passed to the `cleaners()` function
    """
    if "column" not in arguments:
        raise ValueError("`column` is a required argument")
    cleaner_name = f"{function.__name__}_{arguments['column']}"
    return {cleaner_name: {"function": function, "arguments": arguments}}


def cleaners(*cleaner: dict[str, dict[str, Any]]) -> dict[str, dict[str, Any]]:
    """Combine multiple cleaners in a single object to pass to `process()`.

    Args:
        cleaner: Output of the `cleaner()` function

    Returns:
        A representation of multiple cleaners to be passed to the `process()` function

    Examples:
        ```python
        clean_pipeline = cleaners(
            cleaner(
                normalise_company_number,
                {"column": "company_number"},
            ),
            cleaner(
                normalise_postcode,
                {"column": "postcode"},
            ),
        )
        ```
    """
    return {k: v for d in cleaner for k, v in d.items()}


def process(data: DataFrame, pipeline: dict[str, dict[str, Any]]) -> DataFrame:
    """Apply cleaners to input dataframe.

    Args:
        data: The dataframe to process
        pipeline: Output of the `cleaners()` function

    Returns:
        The processed dataset
    """
    curr = data
    for func in pipeline.keys():
        curr = pipeline[func]["function"](curr, **pipeline[func]["arguments"])
    return curr



"""Functions to select and retrieve data from the Matchbox server."""

import itertools
from typing import Generator, Iterator, Literal, get_args

import polars as pl
from polars import DataFrame as PolarsDataFrame
from pydantic import BaseModel, ConfigDict
from sqlalchemy import Engine, create_engine

from matchbox.client import _handler
from matchbox.client._settings import settings
from matchbox.common.db import QueryReturnType, ReturnTypeStr
from matchbox.common.graph import DEFAULT_RESOLUTION
from matchbox.common.logging import logger
from matchbox.common.sources import Match, Source, SourceAddress


class Selector(BaseModel):
    """A selector to choose a source and optionally a subset of columns to select."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    address: SourceAddress
    fields: list[str] | None = None
    engine: Engine


def select(
    *selection: str | dict[str, str],
    engine: Engine | None = None,
) -> list[Selector]:
    """From one engine, builds and verifies a list of selectors.

    Args:
        selection: Full source names and optionally a subset of columns to select
        engine: The engine to connect to the data warehouse hosting the source.
            If not provided, will use a connection string from the
            `MB__CLIENT__DEFAULT_WAREHOUSE` environment variable.

    Returns:
        A list of Selector objects

    Examples:
        ```python
        select("companies_house", "hmrc_exporters", engine=engine)
        ```

        ```python
        select({"companies_house": ["crn"], "hmrc_exporters": ["name"]}, engine=engine)
        ```
    """
    if not engine:
        if default_engine := settings.default_warehouse:
            engine = create_engine(default_engine)
            logger.warning("Using default engine")
        else:
            raise ValueError(
                "An engine needs to be provided if "
                "`MB__CLIENT__DEFAULT_WAREHOUSE` is unset"
            )

    selectors = []
    for s in selection:
        if isinstance(s, str):
            source_address = SourceAddress.compose(engine, s)
            selectors.append(Selector(engine=engine, address=source_address))
        elif isinstance(s, dict):
            for full_name, fields in s.items():
                source_address = SourceAddress.compose(engine, full_name)

                selectors.append(
                    Selector(engine=engine, address=source_address, fields=fields)
                )
        else:
            raise ValueError("Selection specified in incorrect format")

    return selectors


def _process_query_result(
    data: PolarsDataFrame, selector: Selector, mb_ids: PolarsDataFrame, db_pk: str
) -> PolarsDataFrame:
    """Process query results by joining with matchbox IDs and filtering fields.

    Args:
        data: The raw data from the source
        selector: The selector with source and fields information
        mb_ids: The matchbox IDs
        db_pk: The primary key of the source

    Returns:
        The processed table with joined matchbox IDs and filtered fields
    """
    # Join data with matchbox IDs
    joined_table = data.join(
        other=mb_ids,
        left_on=selector.address.format_column(db_pk),
        right_on="source_pk",
        how="inner",
    )

    # Apply field filtering if needed
    if selector.fields:
        keep_cols = ["id"] + [
            selector.address.format_column(f) for f in selector.fields
        ]
        match_cols = [col for col in joined_table.columns if col in keep_cols]
        return joined_table.select(match_cols)
    else:
        return joined_table


def _source_query(
    selector: Selector,
    return_batches: bool = False,
    batch_size: int | None = None,
    only_indexed: bool = False,
) -> tuple[Source, Iterator[PolarsDataFrame]]:
    """From a Selector, query a source and join to matchbox IDs."""
    source = _handler.get_source(selector.address).set_engine(selector.engine)

    indexed_columns = set()
    if source.columns:
        indexed_columns = set([col.name for col in source.columns])

    # If only_indexed is True and source.columns is unset, we will raise
    if only_indexed and selector.fields and not set(selector.fields) <= indexed_columns:
        raise ValueError("Attempting to query unindexed columns.")

    selected_fields = None
    if selector.fields:
        selected_fields = list(set(selector.fields))

    raw_results = source.to_polars(
        fields=selected_fields,
        return_batches=return_batches,
        batch_size=batch_size,
    )

    if isinstance(raw_results, PolarsDataFrame):
        raw_results = [raw_results]

    return source, raw_results


def _process_selectors(
    selectors: list[Selector],
    resolution_name: str | None,
    threshold: int | None,
    batch_size: int | None,
    only_indexed: bool,
) -> Iterator[PolarsDataFrame]:
    """Helper function to process selectors and return an iterator of results.

    For non-batched queries, turn this into a list.

    For batched queries, yield from it.
    """
    # Create iterators for each selector
    selector_iters: list[Generator[PolarsDataFrame, None, None]] = []

    def _process_batches(
        batches: Iterator[PolarsDataFrame],
        selector: Selector,
        mb_ids: PolarsDataFrame,
        db_pk: str,
    ) -> Generator[PolarsDataFrame, None, None]:
        """Process and transform each batch of results."""
        for batch in batches:
            yield _process_query_result(batch, selector, mb_ids, db_pk=db_pk)

    for selector in selectors:
        mb_ids = pl.from_arrow(
            _handler.query(
                source_address=selector.address,
                resolution_name=resolution_name,
                threshold=threshold,
            )
        )

        source, raw_batches = _source_query(
            selector=selector,
            return_batches=True,
            batch_size=batch_size,
            only_indexed=only_indexed,
        )

        # Process and transform each batch
        selector_iters.append(
            _process_batches(
                batches=raw_batches,
                selector=selector,
                mb_ids=mb_ids,
                db_pk=source.db_pk,
            )
        )

    # Chain iterators if multiple selectors
    if len(selector_iters) == 1:
        batches_iter = selector_iters[0]
    else:
        # Interleave batches from different selectors
        batches_iter = itertools.chain.from_iterable(selector_iters)

    return batches_iter


def query(
    *selectors: list[Selector],
    resolution_name: str | None = None,
    combine_type: Literal["concat", "explode", "set_agg"] = "concat",
    return_type: ReturnTypeStr = "pandas",
    threshold: int | None = None,
    batch_size: int | None = None,
    return_batches: bool = False,
    only_indexed: bool = False,
) -> QueryReturnType | Iterator[QueryReturnType]:
    """Runs queries against the selected backend.

    Args:
        selectors: Each selector is the output of `select()`.
            This allows querying sources coming from different engines
        resolution_name (optional): The name of the resolution point to query
            If not set:

            * If querying a single source, it will use the source resolution
            * If querying 2 or more sources, it will look for a default resolution
        combine_type: How to combine the data from different sources.

            * If `concat`, concatenate all sources queried without any merging.
                Multiple rows per ID, with null values where data isn't available
            * If `explode`, outer join on Matchbox ID. Multiple rows per ID,
                with one for every unique combination of data requested
                across all sources
            * If `set_agg`, join on Matchbox ID, group on Matchbox ID, then
                aggregate to nested lists of unique values. One row per ID,
                but all requested data is in nested arrays
        return_type: The form to return data in, one of "pandas" or "arrow"
            Defaults to pandas for ease of use
        threshold (optional): The threshold to use for creating clusters
            If None, uses the resolutions' default threshold
            If an integer, uses that threshold for the specified resolution, and the
            resolution's cached thresholds for its ancestors
        batch_size (optional): The size of each batch when fetching data from the
            warehouse, which helps reduce memory usage and load on the database.
            Default is None.
        return_batches (optional): If True, returns an iterator of batches instead of a
            single combined result, which is useful for processing large datasets with
            limited memory. Default is False.
        only_indexed (optional): If True, it will raise an exception when attempting to
            query un-indexed columns, which should never be done if querying for
            subsequent matching. Default is False.

    Returns:
        If return_batches is False:
            Data in the requested return type (DataFrame or ArrowTable)
        If return_batches is True:
            An iterator yielding batches in the requested return type

    Examples:
        ```python
        query(
            select({"companies_house": ["crn", "name"]}, engine=engine),
        )
        ```

        ```python
        query(
            select("companies_house", engine=engine1),
            select("datahub_companies", engine=engine2),
            resolution_name="last_linker",
        )
        ```

        ```python
        # Process large results in batches of 5000 rows
        for batch in query(
            select("companies_house", engine=engine),
            batch_size=5000,
            return_batches=True,
        ):
            batch.head()
        ```
    """
    # Validate arguments
    if combine_type not in ("concat", "explode", "set_agg"):
        raise ValueError(f"combine_type of {combine_type} not valid")

    if return_type not in get_args(ReturnTypeStr):
        raise ValueError(f"return_type of {return_type} not valid")

    if not selectors:
        raise ValueError("At least one selector must be specified")

    if return_batches and combine_type != "concat":
        raise ValueError("Batching is only supported for `combine_type='concat'`")

    selectors: list[Selector] = list(itertools.chain(*selectors))

    if not resolution_name and len(selectors) > 1:
        resolution_name = DEFAULT_RESOLUTION

    res = _process_selectors(
        selectors=selectors,
        resolution_name=resolution_name,
        threshold=threshold,
        batch_size=batch_size,
        only_indexed=only_indexed,
    )
    if return_batches:
        # Return an iterator of batches
        def generate_batches() -> Iterator[Generator[PolarsDataFrame, None, None]]:
            """Yield batches of data in the requested format."""
            for batch in res:
                match return_type:
                    case "pandas":
                        yield batch.to_pandas()
                    case "polars":
                        yield batch
                    case "arrow":
                        yield batch.to_arrow()

        return generate_batches()

    else:
        # Process all data and return a single result
        tables: list[PolarsDataFrame] = list(res)

        # Make sure we have some results
        if not tables:
            result = pl.DataFrame()
        else:
            # Combine results based on combine_type
            if combine_type == "concat":
                result = pl.concat(tables, how="diagonal")
            else:
                result = tables[0]
                for table in tables[1:]:
                    result = result.join(table, on="id", how="full", coalesce=True)

                result = result.select(["id", pl.all().exclude("id")])

                if combine_type == "set_agg":
                    # Aggregate into lists
                    agg_expressions = [
                        pl.col(col).unique() for col in result.columns if col != "id"
                    ]
                    result = result.group_by("id").agg(agg_expressions)

        # Return in requested format
        match return_type:
            case "pandas":
                return result.to_pandas()
            case "polars":
                return result
            case "arrow":
                return result.to_arrow()


def match(
    *targets: list[Selector],
    source: list[Selector],
    source_pk: str,
    resolution_name: str = DEFAULT_RESOLUTION,
    threshold: int | None = None,
) -> list[Match]:
    """Matches IDs against the selected backend.

    Args:
        targets: Each target is the output of `select()`.
            This allows matching against sources coming from different engines
        source: The output of using `select()` on a single source.
        source_pk: The primary key value to match from the source.
        resolution_name (optional): The resolution name to use for filtering results.
            If not set, it will look for a default resolution.
        threshold (optional): The threshold to use for creating clusters.
            If None, uses the resolutions' default threshold
            If an integer, uses that threshold for the specified resolution, and the
            resolution's cached thresholds for its ancestors

    Examples:
        ```python
        mb.match(
            select("datahub_companies", engine=engine),
            source=select("companies_house", engine=engine),
            source_pk="8534735",
            resolution_name="last_linker",
        )
        ```
    """
    if len(source) > 1:
        raise ValueError("Only one source can be matched at one time")
    source = source[0].address

    targets: list[Selector] = list(itertools.chain(*targets))
    targets = [t.address for t in targets]

    return _handler.match(
        targets=targets,
        source=source,
        source_pk=source_pk,
        resolution_name=resolution_name,
        threshold=threshold,
    )



"""Logging utilities."""

import datetime
import json
import logging
from functools import cached_property
from importlib.metadata import version
from typing import Any, Final, Literal

from rich.console import Console
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

LogLevelType = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
"""Type for all Python log levels."""


class PrefixedLoggerAdapter(logging.LoggerAdapter):
    """A logger adapter that supports adding a prefix enclosed in square brackets.

    This adapter allows passing an optional prefix parameter to any logging call
    without modifying the underlying logger.
    """

    def process(self, msg: Any, kwargs: dict[str, Any]) -> tuple[Any, dict[str, Any]]:
        """Process the log message, adding a prefix if provided.

        Args:
            msg: The log message
            kwargs: Additional arguments to the logging method

        Returns:
            Tuple of (modified_message, modified_kwargs)
        """
        prefix = kwargs.pop("prefix", None)

        if prefix:
            msg = f"[{prefix}] {msg}"

        return msg, kwargs


logger: Final[PrefixedLoggerAdapter]
"""Logger for Matchbox.

Used for all logging in the Matchbox library.

Allows passing a prefix to any logging call.

Examples:
    ```python
    log_prefix = f"Model {name}"
    logger.debug("Inserting metadata", prefix=log_prefix)
    logger.debug("Inserting data", prefix=log_prefix)
    logger.info("Insert successful", prefix=log_prefix)
    ```
"""
logger = PrefixedLoggerAdapter(logging.getLogger("matchbox"), {})


console: Final[Console] = Console()
"""Console for Matchbox.

Used for any CLI utilities in the Matchbox library.
"""


def build_progress_bar(console_: Console | None = None) -> Progress:
    """Create a progress bar."""
    if console_ is None:
        console_ = console

    return Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
        console=console_,
    )


class ASIMFormatter(logging.Formatter):
    """Format logging with ASIM standard fields."""

    @cached_property
    def matchbox_version(self) -> str:
        """Cached matchbox version."""
        return version("matchbox_db")

    @cached_property
    def event_severity(self) -> dict[str, str]:
        """Event severity level lookup."""
        return {
            "DEBUG": "Informational",
            "INFO": "Informational",
            "WARNING": "Low",
            "ERROR": "Medium",
            "CRITICAL": "High",
        }

    def format(self, record) -> str:
        """Convert logs to JSON."""
        log_time = datetime.datetime.utcfromtimestamp(record.created).isoformat()
        return json.dumps(
            {
                "EventMessage": record.getMessage(),
                "EventCount": 1,
                "EventStartTime": log_time,
                "EventEndTime": log_time,
                "EventType": record.name,
                "EventResult": "NA",
                "EventSeverity": self.event_severity[record.levelname],
                "EventOriginalSeverity": record.levelname,
                "EventSchema": "ProcessEvent",
                "EventSchemaVersion": "0.1.4",
                "EventVendor": "Matchbox",
                "EventProduct": "Matchbox",
                "AdditionalFields": {
                    "MatchboxVersion": self.matchbox_version,
                },
            }
        )



import polars as pl
import pyarrow as pa
from httpx import Response
from polars.testing import assert_frame_equal
from respx import MockRouter
from sqlalchemy import Engine, create_engine, text

from matchbox.client.extract import sql_interface
from matchbox.common.arrow import SCHEMA_MB_IDS, table_to_buffer
from matchbox.common.factories.sources import source_from_tuple


def test_sql_interface(
    sqlite_warehouse: Engine,
    matchbox_api: MockRouter,
):
    # Make dummy data
    foo = source_from_tuple(
        full_name="foo",
        engine=sqlite_warehouse,
        data_pks=[1, 2, 3],
        data_tuple=({"col": 0}, {"col": 1}, {"col": 2}),
    )
    bar = source_from_tuple(
        full_name="bar",
        engine=sqlite_warehouse,
        data_pks=["a", "b", "c"],
        data_tuple=({"col": 10}, {"col": 11}, {"col": 12}),
    )
    different_engine_source = source_from_tuple(
        full_name="baz",
        engine=create_engine("sqlite:///:memory:"),
        data_pks=["a", "b", "c"],
        data_tuple=({"col": 10}, {"col": 11}, {"col": 12}),
    )

    foo.to_warehouse(sqlite_warehouse)
    bar.to_warehouse(sqlite_warehouse)

    # Because of FULL OUTER JOIN, we expect some values to be null, and some explosions
    df_expected = pl.DataFrame(
        [
            {"foo_pk": 1, "bar_pk": "a", "foo_col": 0, "bar_col": 10},
            {"foo_pk": 2, "bar_pk": None, "foo_col": 1, "bar_col": None},
            {"foo_pk": 3, "bar_pk": "b", "foo_col": 2, "bar_col": 11},
            {"foo_pk": 3, "bar_pk": "c", "foo_col": 2, "bar_col": 12},
        ]
    )

    # Mock API
    matchbox_api.get("/sources").mock(
        return_value=Response(
            200,
            json=[
                foo.source.model_dump(),
                bar.source.model_dump(),
                different_engine_source.source.model_dump(),
            ],
        )
    )

    # We don't expect a query call for the third source
    matchbox_api.get("/query", params={"full_name": "foo"}).mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"source_pk": "1", "id": 1},
                        {"source_pk": "2", "id": 2},
                        {"source_pk": "3", "id": 3},
                    ],
                    schema=SCHEMA_MB_IDS,
                )
            ).read(),
        )
    )

    matchbox_api.get("/query", params={"full_name": "bar"}).mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"source_pk": "a", "id": 1},
                        {"source_pk": "b", "id": 3},
                        {"source_pk": "c", "id": 3},
                    ],
                    schema=SCHEMA_MB_IDS,
                )
            ).read(),
        )
    )

    # Call extracting method
    mapping, column_names, view_definition = sql_interface(
        resolution_name="companies",
        engine=sqlite_warehouse,
        mapping_table="companies_matches",
    )

    # Check column names are as expected
    col_n_df = column_names.to_pandas()
    assert set(col_n_df["table_name"].tolist()) == {"foo", "bar"}
    assert set(col_n_df["column_name"].tolist()) == {"col", "pk"}
    assert set(col_n_df["db_pk"].tolist()) == {"pk"}
    assert set(col_n_df["combined_name"].tolist()) == {
        "foo_col",
        "bar_col",
        "foo_pk",
        "bar_pk",
    }

    # Write results to the warehouse
    with sqlite_warehouse.connect() as conn:
        pl.from_arrow(mapping).write_database("companies_matches", conn)
        pl.from_arrow(column_names).write_database("companies_columns", conn)
        conn.execute(text(f"CREATE VIEW companies AS {view_definition}"))
        conn.commit()

        # Check SQL interface works as intended
        test_query = "SELECT * FROM companies"
        df_actual = pl.read_database(test_query, connection=conn)

        assert_frame_equal(
            df_actual, df_expected, check_row_order=False, check_column_order=False
        )



"""Implementation of default cleaning functions."""

from functools import partial

from pandas import DataFrame

from matchbox.client.clean import steps
from matchbox.client.clean import utils as cu


def company_name(
    df: DataFrame,
    column: str,
    column_secondary: str = None,
    stopwords: str = cu.STOPWORDS,
) -> DataFrame:
    """Standard cleaning function for company names.

    * Lower case, remove punctuation & tokenise the company name into an array
    * Extract tokens into: 'unusual' and 'stopwords'. Dedupe. Sort alphabetically
    * Untokenise the unusual words back to a string

    Args:
        df: a dataframe
        column: a column containing the company's main name
        column_secondary: a column containing an array of the company's
            secondary names
        stopwords: a list of stopwords to use for this clean

    Returns:
        dataframe: the same as went in, but cleaned
    """
    remove_stopwords = partial(steps.remove_stopwords, stopwords=stopwords)

    clean_primary = cu.cleaning_function(
        steps.clean_punctuation,
        steps.expand_abbreviations,
        steps.tokenise,  # returns array
        remove_stopwords,
        steps.list_join_to_string,  # returns col
    )

    clean_secondary = cu.unnest_renest(clean_primary)

    df = clean_primary(df, column)

    if column_secondary is not None:
        df = clean_secondary(df, column_secondary)

    return df


def company_number(df: DataFrame, column: str) -> DataFrame:
    """Remove non-numbers, and then leading zeroes.

    Args:
        df: a dataframe
        column: a column containing a company number

    Returns:
        dataframe: the same as went in, but cleaned
    """
    clean_number = cu.cleaning_function(steps.remove_notnumbers_leadingzeroes)

    df = clean_number(df, column)

    return df


def postcode(df: DataFrame, column: str) -> DataFrame:
    """Removes all punctuation, converts to upper, removes all spaces.

    Args:
        df: a dataframe
        column: a column containing a postcode

    Returns:
        dataframe: the same as went in, but cleaned

    """
    clean_postcode = cu.cleaning_function(
        steps.punctuation_to_spaces, steps.to_upper, steps.remove_whitespace
    )

    df = clean_postcode(df, column)

    return df


def postcode_to_area(df: DataFrame, column: str) -> DataFrame:
    """Extracts postcode area from a postcode.

    Args:
        df: a dataframe
        column: a column containing a postcode

    Returns:
        dataframe: the same as went in, but cleaned
    """
    extract_area = cu.cleaning_function(steps.get_postcode_area)

    df = extract_area(df, column)

    return df


def extract_company_number_to_new(
    df: DataFrame, column: str, new_column: str
) -> DataFrame:
    """Detects the Companies House CRN in a column and moves it to a new column.

    Args:
        df: a dataframe
        column: a column containing some company numbers
        new_column: the name of the column to add

    Returns:
        dataframe: the same as went in with a new column for CRNs
    """
    clean_crn = cu.cleaning_function(
        steps.clean_punctuation_except_hyphens,
        steps.to_upper,
        steps.filter_company_number,
    )

    clean_crn_aliased = cu.alias(clean_crn, alias=new_column)

    df = clean_crn_aliased(df, column)

    return df


def extract_duns_number_to_new(
    df: DataFrame, column: str, new_column: str
) -> DataFrame:
    """Detects the Dun & Bradstreet DUNS nuber in a column and moves it to a new column.

    Args:
        df: a dataframe
        column: a column containing some DUNS numbers
        new_column: the name of the column to add

    Returns:
        dataframe: the same as went in with a new column for DUNs numbers
    """
    clean_duns = cu.cleaning_function(
        steps.clean_punctuation_except_hyphens, steps.to_upper, steps.filter_duns_number
    )

    clean_duns_aliased = cu.alias(clean_duns, alias=new_column)

    df = clean_duns_aliased(df, column)

    return df


def extract_cdms_number_to_new(
    df: DataFrame, column: str, new_column: str
) -> DataFrame:
    """Detects the CDMS nuber in a column and moves it to a new column.

    Args:
        df: a dataframe
        column: a column containing some CDMS numbers
        new_column: the name of the column to add

    Returns:
        dataframe: the same as went in with a new column for CDMS numbers
    """
    clean_cdms = cu.cleaning_function(
        steps.clean_punctuation_except_hyphens, steps.to_upper, steps.filter_cdms_number
    )

    clean_cdms_aliased = cu.alias(clean_cdms, alias=new_column)

    df = clean_cdms_aliased(df, column)

    return df


def drop(df: DataFrame, column: str) -> DataFrame:
    """Drops the column from the dataframe.

    Args:
        df: a dataframe
        column: a column

    Returns:
        dataframe: the same as went in without the column
    """
    return df.drop(columns=[column])



"""Functions to extract data out of the Matchbox server."""

from pyarrow import Table as ArrowTable
from sqlalchemy import Engine, String, func, select, text

from matchbox.client import _handler
from matchbox.common.sources import Source, SourceAddress


def _create_view_definition(
    column_names: list[dict[str, str]],
    db_pks: dict[str, str],
    mapping_table: str,
    engine: Engine,
) -> str:
    selection = []

    for cn in column_names:
        selection.append(
            text(f"{cn['table_name']}.{cn['column_name']} as {cn['combined_name']}")
        )

    query = select(*selection).select_from(text(mapping_table))

    for table_name, db_pk in db_pks.items():
        query = query.join(
            text(table_name),
            func.cast(text(f"{table_name}.{db_pk}"), String)
            == func.cast(text(f"{mapping_table}.{table_name}_{db_pk}"), String),
            isouter=True,
        )

    return str(query.compile(dialect=engine.dialect))


def _combined_colname(source: Source, col_name: str):
    return source.address.full_name.replace(".", "_") + "_" + col_name


def sql_interface(
    resolution_name: str,
    engine: Engine,
    mapping_table: str,
) -> tuple[ArrowTable, ArrowTable, str]:
    """Generate snapshot and DDL implmenting SQL interface to results.

    Args:
        resolution_name: Name of the resolution from which to generate results
        engine: SQLAlchemy engine of warehouse running SQL interface
        mapping_table: The name (+ schema) where the mapping table will be written.

    Returns:
        A tuple of 3 items containing:

            * A PyArrow table mapping from Matchbox IDs to all source PKs
            * A PyArrow table mapping sources and columns to column names in the view
            * The DQL definition of a view to use for querying results in SQL

    """
    # Get all sources in scope of the resolution
    res_sources = _handler.get_resolution_sources(resolution_name=resolution_name)

    # Filter only sources compatible with engine
    warehouse_hash_b64 = SourceAddress.compose(
        full_name="", engine=engine
    ).warehouse_hash_b64

    sources = [
        s.set_engine(engine)
        for s in res_sources
        if s.address.warehouse_hash_b64 == warehouse_hash_b64
    ]

    source_mb_ids: list[ArrowTable] = []
    column_names: list[dict[str, str]] = []
    db_pks: dict[str, str] = {}

    for s in sources:
        # Get Matchbox IDs from backend
        source_mb_ids.append(
            _handler.query(
                source_address=s.address,
                resolution_name=resolution_name,
            )
        )
        # Get remote columns from warehouse
        cols = s.get_remote_columns()
        column_names += [
            {
                "table_name": s.address.full_name,
                "column_name": c,
                "combined_name": _combined_colname(s, c),
                "db_pk": s.db_pk,
            }
            for c in cols
        ]

        db_pks[s.address.full_name] = s.db_pk

    # Join Matchbox IDs to form mapping table
    mapping = source_mb_ids[0]
    mapping = mapping.rename_columns(
        {
            "source_pk": _combined_colname(
                sources[0], db_pks[sources[0].address.full_name]
            )
        }
    )
    for s, mb_ids in zip(sources[1:], source_mb_ids[1:], strict=True):
        mapping = mapping.join(right_table=mb_ids, keys="id", join_type="full outer")
        mapping = mapping.rename_columns(
            {"source_pk": _combined_colname(s, db_pks[s.address.full_name])}
        )

    # Get SQL query for view
    view_definition = _create_view_definition(
        column_names=column_names,
        db_pks=db_pks,
        mapping_table=mapping_table,
        engine=engine,
    )

    # Convert column names dataframe
    column_names = ArrowTable.from_pylist(column_names)

    return mapping, column_names, view_definition



import logging
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator
from unittest.mock import patch

import pytest
from rich.console import Console
from rich.progress import Progress

pytest_plugins = [
    "test.fixtures.db",
    "test.fixtures.graph",
    "test.fixtures.client",
]

TEST_ROOT = Path(__file__).resolve().parent


def pytest_configure():
    """Configure pytest settings."""

    # Quieten down the logging for specific loggers
    logging.getLogger("faker").setLevel(logging.WARNING)


@pytest.fixture(scope="session")
def test_root_dir() -> Path:
    return TEST_ROOT


@contextmanager
def parallel_pool_for_tests(
    max_workers: int = 2, timeout: int = 30
) -> Iterator[ThreadPoolExecutor]:
    """Context manager for safe parallel execution in tests using threads.

    Args:
        max_workers: Maximum number of worker threads
        timeout: Maximum seconds to wait for each task
    """
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        try:
            yield executor
        finally:
            executor.shutdown(wait=False, cancel_futures=True)


@pytest.fixture(scope="session", autouse=True)
def patch_multiprocessing() -> Iterator[None]:
    """Patch ProcessPoolExecutor to use ThreadPoolExecutor in tests."""
    with patch(
        "matchbox.common.transform.ProcessPoolExecutor",
        lambda *args, **kwargs: parallel_pool_for_tests(timeout=30),
    ):
        yield


@pytest.fixture(scope="session", autouse=True)
def patch_rich_console() -> Iterator[None]:
    """Patch Rich console for quiet output in tests."""
    quiet_console = Console(quiet=True)

    console_patch = patch("matchbox.common.logging.console", new=quiet_console)
    progress_patch = patch(
        "matchbox.common.logging.build_progress_bar",
        return_value=Progress(console=quiet_console),
    )

    with console_patch, progress_patch:
        yield



"""Functions to index data sources to the Matchbox server."""

from sqlalchemy import Engine

from matchbox.client import _handler
from matchbox.common.sources import Source, SourceAddress, SourceColumn


def _process_columns(
    columns: list[str] | list[dict[str, dict[str, str]]] | None,
) -> tuple[SourceColumn]:
    if columns is None:
        return []

    if isinstance(columns[0], str):
        return (SourceColumn(name=column) for column in columns)

    return (
        SourceColumn(name=column["name"], type=column["type"]) for column in columns
    )


def index(
    full_name: str,
    db_pk: str,
    engine: Engine,
    resolution_name: str | None = None,
    columns: list[str] | list[dict[str, dict[str, str]]] | None = None,
    batch_size: int | None = None,
) -> None:
    """Indexes data in Matchbox.

    Args:
        full_name: the full name of the source
        db_pk: the primary key of the source
        engine: the engine to connect to a data warehouse
        resolution_name: a custom resolution name
            If missing, will use the default name for a `Source`
        columns: the columns to index
        batch_size: the size of each batch when fetching data from the warehouse,
            which helps reduce the load on the database. Default is None.

    Examples:
        ```python
        index("mb.test_orig", "id", engine=engine)
        ```
        ```python
        index("mb.test_cl2", "id", engine=engine, columns=["name", "age"])
        ```
        ```python
        index(
            "mb.test_cl2",
            "id",
            engine=engine,
            columns=[
                {"name": "name", "type": "TEXT"},
                {"name": "age", "type": "BIGINT"},
            ],
        )
        ```
        ```python
        index("mb.test_orig", "id", engine=engine, batch_size=10_000)
        ```
    """
    columns = _process_columns(columns)

    address = SourceAddress.compose(engine=engine, full_name=full_name)
    if resolution_name:
        source = Source(
            address=address,
            resolution_name=resolution_name,
            columns=columns,
            db_pk=db_pk,
        )
    else:
        source = Source(
            address=address,
            columns=columns,
            db_pk=db_pk,
        )

    source.set_engine(engine)
    if not columns:
        source = source.default_columns()

    _handler.index(source=source, batch_size=batch_size)



"""Script to benchmark the full dummy data generation pipeline."""

import time
from contextlib import contextmanager
from pathlib import Path

import pyarrow as pa

from matchbox.common.factories.models import generate_dummy_probabilities
from matchbox.common.hash import hash_data, hash_values
from matchbox.common.logging import console
from matchbox.common.transform import (
    attach_components_to_probabilities,
    to_hierarchical_clusters,
)
from matchbox.server.postgresql.benchmark.generate_tables import PRESETS
from matchbox.server.postgresql.utils.insert import HashIDMap

ROOT = Path(__file__).parent.parent


@contextmanager
def timer(description: str):
    """Context manager to time a block of code."""
    start = time.time()
    yield
    elapsed = time.time() - start

    if elapsed >= 60:
        minutes = int(elapsed // 60)
        seconds = elapsed % 60
        time_str = f"{minutes} min {seconds:.1f} sec"
    else:
        time_str = f"{elapsed:.2f} seconds"

    console.log(f"{description} in {time_str}")


if __name__ == "__main__":
    config = PRESETS["l"]
    left_ids = tuple(range(config["dedupe_components"]))
    right_ids = tuple(
        range(config["dedupe_components"], config["dedupe_components"] * 2)
    )
    probs = generate_dummy_probabilities(
        left_ids, right_ids, [0.6, 1], config["link_components"], config["link_len"]
    )

    all_probs = pa.concat_arrays(
        [probs["left_id"].combine_chunks(), probs["right_id"].combine_chunks()]
    )

    lookup = pa.table(
        {
            "id": all_probs,
            "hash": pa.array(
                [hash_data(p) for p in all_probs.to_pylist()], type=pa.large_binary()
            ),
        }
    )

    hm = HashIDMap(start=max(right_ids) + 1, lookup=lookup)

    with timer("Full pipeline completed"):
        console.log(f"Processing {len(probs):,} records")

        with timer("Added components"):
            probs_with_ccs = attach_components_to_probabilities(
                pa.table(
                    {
                        "left_id": hm.get_hashes(probs["left_id"]),
                        "right_id": hm.get_hashes(probs["right_id"]),
                        "probability": probs["probability"],
                    }
                )
            )

        with timer("Built hierarchical clusters"):
            hierarchy = to_hierarchical_clusters(
                probabilities=probs_with_ccs,
                hash_func=hash_values,
                dtype=pa.large_binary,
            )



import uuid

import polars as pl
import pyarrow as pa
import pytest

from matchbox.common.hash import HashMethod, IntMap, hash_arrow_table, hash_rows


def test_intmap_basic():
    im1 = IntMap(salt=10)
    a = im1.index(1, 2)
    b = im1.index(3, 4)
    c = im1.index(a, b)

    assert len({a, b, c}) == 3
    assert max(a, b, c) < 0


def test_intmap_same():
    im1 = IntMap(salt=10)
    a = im1.index(1, 2)
    b = im1.index(3, 4)
    c = im1.index(a, b)

    im2 = IntMap(salt=10)
    x = im2.index(1, 2)
    y = im2.index(3, 4)
    z = im2.index(a, b)

    assert (a, b, c) == (x, y, z)


def test_intmap_different():
    im1 = IntMap(salt=10)
    a = im1.index(1, 2)
    b = im1.index(3, 4)
    c = im1.index(a, b)

    im2 = IntMap(salt=11)
    x = im2.index(1, 2)
    y = im2.index(3, 4)
    z = im2.index(a, b)

    for v1, v2 in zip([a, b, c], [x, y, z], strict=True):
        assert v1 != v2


def test_intmap_unordered():
    im1 = IntMap(salt=10)
    a = im1.index(1, 2, 3)
    b = im1.index(3, 1, 2)

    assert a == b


@pytest.mark.parametrize(
    ["method"],
    [
        pytest.param(HashMethod.SHA256, id="sha256"),
        pytest.param(HashMethod.XXH3_128, id="xxh3_128"),
    ],
)
def test_hash_rows(method: HashMethod):
    data = pl.DataFrame(
        {
            "string_col": ["abc", "def", "ghi"],
            "int_col": [1, 2, 3],
            "float_col": [1.1, 2.2, 3.3],
            "struct_col": [{"a": 1, "b": "x"}, {"a": 2, "b": "y"}, {"a": 3, "b": "z"}],
            "object_col": [uuid.uuid4(), uuid.uuid4(), uuid.uuid4()],
            "binary_col": [b"data1", b"data2", b"data3"],
        }
    )

    assert isinstance(data["string_col"].dtype, pl.String)
    assert isinstance(data["int_col"].dtype, pl.Int64)
    assert isinstance(data["float_col"].dtype, pl.Float64)
    assert isinstance(data["struct_col"].dtype, pl.Struct)
    assert isinstance(data["object_col"].dtype, pl.Object)
    assert isinstance(data["binary_col"].dtype, pl.Binary)

    hash_rows(data, columns=data.columns, method=method)


@pytest.mark.parametrize(
    ["method"],
    [
        pytest.param(HashMethod.SHA256, id="sha256"),
        pytest.param(HashMethod.XXH3_128, id="xxh3_128"),
    ],
)
def test_hash_arrow_table(method: HashMethod):
    a = pa.Table.from_pydict(
        {
            "a": [1, 2, 3],
            "b": [4, 5, 6],
        }
    )
    # Column order should not matter
    b = pa.Table.from_pydict(
        {
            "b": [4, 5, 6],
            "a": [1, 2, 3],
        }
    )
    # Row order should not matter
    c = pa.Table.from_pydict(
        {
            "a": [3, 2, 1],
            "b": [6, 5, 4],
        }
    )
    # Column and row order should not matter
    d = pa.Table.from_pydict(
        {
            "b": [6, 5, 4],
            "a": [3, 2, 1],
        }
    )
    # Empty table should have a different hash
    e = pa.Table.from_pydict(
        {
            "a": [],
            "b": [],
        }
    )
    # Different row data should have a different hash
    f = pa.Table.from_pydict(
        {
            "a": [1, 2, 3],
            "b": [4, 5, 7],
        }
    )
    # If column name change their order, the hash should change
    g = pa.Table.from_pydict(
        {
            "b": [1, 2, 3],
            "a": [4, 5, 6],
        }
    )
    # List columns are handled
    h = pa.Table.from_pydict(
        {
            "a": [1, 2, 3],
            "b": [[1, 2], [3, 4], [5, 6]],
        }
    )
    # List order doesn't matter
    i = pa.Table.from_pydict(
        {
            "a": [1, 2, 3],
            "b": [[2, 1], [4, 3], [6, 5]],
        }
    )
    # Binary columns are handled, including non-UTF-8 bytes
    j = pa.Table.from_pydict(
        {
            "a": [1, 2, 3],
            "b": [b"abc", None, bytes([255, 254, 253])],
        }
    )

    h_a = hash_arrow_table(a, method=method)
    h_a1 = hash_arrow_table(a, method=method)
    h_b = hash_arrow_table(b, method=method)
    h_c = hash_arrow_table(c, method=method)
    h_d = hash_arrow_table(d, method=method)
    h_e = hash_arrow_table(e, method=method)
    h_f = hash_arrow_table(f, method=method)
    h_g = hash_arrow_table(g, method=method)
    h_h = hash_arrow_table(h, method=method)
    h_i = hash_arrow_table(i, method=method)
    h_j = hash_arrow_table(j, method=method)

    # Basic type check
    assert isinstance(h_a, bytes)
    # Basic invariance checks
    assert h_a == h_a1 == h_b == h_c == h_d
    # Different data = different hashes
    assert h_a != h_e
    assert h_a != h_f
    assert h_a != h_g
    assert h_a != h_j
    # List type table should be consistent regardless of column order
    assert h_h == h_i


@pytest.mark.parametrize(
    ["method"],
    [
        pytest.param(HashMethod.SHA256, id="sha256"),
        pytest.param(HashMethod.XXH3_128, id="xxh3_128"),
    ],
)
def test_struct_json_hashing(method: HashMethod):
    """Test that struct/JSON data can be properly hashed."""

    # Basic struct test
    a = pa.Table.from_pydict(
        {
            "id": [1, 2, 3],
            "metadata": [
                {"name": "Alice", "age": 30},
                {"name": "Bob", "age": 25},
                {"name": "Charlie", "age": 35},
            ],
        }
    )

    assert isinstance(pl.from_arrow(a)["metadata"].dtype, pl.Struct)

    # Same data but different struct serialization
    b = pa.Table.from_pydict(
        {
            "id": [1, 2, 3],
            "metadata": [
                {"age": 30, "name": "Alice"},
                {"age": 25, "name": "Bob"},
                {"age": 35, "name": "Charlie"},
            ],
        }
    )

    # Different data in structs
    c = pa.Table.from_pydict(
        {
            "id": [1, 2, 3],
            "metadata": [
                {"name": "Alice", "age": 31},  # Changed age
                {"name": "Bob", "age": 25},
                {"name": "Charlie", "age": 35},
            ],
        }
    )

    # Nested structs
    d = pa.Table.from_pydict(
        {
            "id": [1, 2, 3],
            "metadata": [
                {"name": "Alice", "details": {"city": "New York", "active": True}},
                {"name": "Bob", "details": {"city": "Boston", "active": False}},
                {"name": "Charlie", "details": {"city": "Chicago", "active": True}},
            ],
        }
    )

    # Test basic struct hashing
    h_a = hash_arrow_table(a, method=method)
    h_a1 = hash_arrow_table(a, method=method)
    h_b = hash_arrow_table(b, method=method)
    h_c = hash_arrow_table(c, method=method)
    h_d = hash_arrow_table(d, method=method)

    # Basic type check
    assert isinstance(h_a, bytes)

    # Basic equality check
    assert h_a == h_a1 == h_b
    # Difference checks
    assert h_a != h_c
    assert h_a != h_d



from functools import lru_cache
from itertools import chain
from typing import Any

import pyarrow as pa
import pyarrow.compute as pc
import pytest

from matchbox.common.factories.models import generate_dummy_probabilities
from matchbox.common.transform import (
    attach_components_to_probabilities,
    component_to_hierarchy,
    to_hierarchical_clusters,
)


@lru_cache(maxsize=None)
def _combine_strings(*n: str) -> str:
    """
    Combine n strings into a single string, with a cache.
    Meant to replace `matchbox.common.hash.IntMap.index`

    Args:
        *args: Variable number of strings to combine

    Returns:
        A single string
    """
    letters = set(chain.from_iterable(n))
    return "".join(sorted(letters))


@pytest.mark.parametrize(
    ("parameters"),
    [
        {
            "left_range": (0, 1_000),
            "right_range": (1_000, 2_000),
            "prob_range": (0.6, 0.8),
            "num_components": 10,
            "total_rows": 100_000,
        },
    ],
    ids=["simple"],
)
def test_attach_components_to_probabilities(parameters: dict[str, Any]):
    left_values = tuple(range(*parameters["left_range"]))
    right_values = tuple(range(*parameters["right_range"]))

    probabilities = generate_dummy_probabilities(
        left_values=left_values,
        right_values=right_values,
        prob_range=parameters["prob_range"],
        num_components=parameters["num_components"],
        total_rows=parameters["total_rows"],
    )

    with_components = attach_components_to_probabilities(probabilities=probabilities)

    assert len(pc.unique(with_components["component"])) == parameters["num_components"]


def test_empty_attach_components_to_probabilities():
    probabilities = pa.table(
        {
            "left_id": [],
            "right_id": [],
            "probability": [],
        }
    )

    with_components = attach_components_to_probabilities(probabilities=probabilities)

    assert len(with_components) == 0


@pytest.mark.parametrize(
    ("probabilities", "hierarchy"),
    [
        # Test case 1: Equal probabilities
        (
            {
                "left_id": ["a", "b", "c"],
                "right_id": ["b", "c", "d"],
                "probability": [100, 100, 100],
            },
            {
                ("ab", "a", 100),
                ("ab", "b", 100),
                ("bc", "b", 100),
                ("bc", "c", 100),
                ("cd", "c", 100),
                ("cd", "d", 100),
                ("abcd", "ab", 100),
                ("abcd", "bc", 100),
                ("abcd", "cd", 100),
            },
        ),
        # Test case 2: Asymmetric probabilities
        (
            {
                "left_id": ["w", "x", "y"],
                "right_id": ["x", "y", "z"],
                "probability": [90, 85, 80],
            },
            {
                ("wx", "w", 90),
                ("wx", "x", 90),
                ("xy", "x", 85),
                ("xy", "y", 85),
                ("wxy", "wx", 85),
                ("wxy", "xy", 85),
                ("yz", "y", 80),
                ("yz", "z", 80),
                ("wxyz", "wxy", 80),
                ("wxyz", "yz", 80),
            },
        ),
        # Test case 3: Single two-item component
        (
            {
                "left_id": ["x"],
                "right_id": ["y"],
                "probability": [90],
            },
            {
                ("xy", "x", 90),
                ("xy", "y", 90),
            },
        ),
        # Test case 4: A component larger than two remains unchanged
        # at a successive threshold
        (
            {
                "left_id": ["x", "y", "a"],
                "right_id": ["y", "z", "b"],
                "probability": [90, 90, 85],
            },
            {
                ("xy", "x", 90),
                ("xy", "y", 90),
                ("yz", "y", 90),
                ("yz", "z", 90),
                ("xyz", "xy", 90),
                ("xyz", "yz", 90),
                ("ab", "a", 85),
                ("ab", "b", 85),
            },
        ),
    ],
    ids=["equal", "asymmetric", "single", "unchanged"],
)
def test_component_to_hierarchy(
    probabilities: dict[str, list[str | float]], hierarchy: set[tuple[str, str, int]]
):
    probabilities_table = (
        pa.Table.from_pydict(probabilities)
        .cast(
            pa.schema(
                [
                    ("left_id", pa.string()),
                    ("right_id", pa.string()),
                    ("probability", pa.uint8()),
                ]
            )
        )
        .sort_by([("probability", "descending")])
    )

    parents, children, probs = zip(*hierarchy, strict=False)

    hierarchy_true = (
        pa.table([parents, children, probs], names=["parent", "child", "probability"])
        .cast(
            pa.schema(
                [
                    ("parent", pa.string()),
                    ("child", pa.string()),
                    ("probability", pa.uint8()),
                ]
            )
        )
        .sort_by(
            [
                ("probability", "descending"),
                ("parent", "ascending"),
                ("child", "ascending"),
            ]
        )
        .filter(pc.is_valid(pc.field("parent")))
    )

    hierarchy = component_to_hierarchy(
        table=probabilities_table, dtype=pa.string, hash_func=_combine_strings
    ).sort_by(
        [
            ("probability", "descending"),
            ("parent", "ascending"),
            ("child", "ascending"),
        ]
    )

    assert hierarchy.equals(hierarchy_true)


@pytest.mark.parametrize(
    ("input_data", "expected_hierarchy"),
    [
        # Single component test case
        (
            {
                "component": [1, 1, 1],
                "left_id": ["a", "b", "c"],
                "right_id": ["b", "c", "d"],
                "probability": [90, 85, 80],
            },
            {
                ("ab", "a", 90),
                ("ab", "b", 90),
                ("bc", "b", 85),
                ("bc", "c", 85),
                ("abc", "ab", 85),
                ("abc", "bc", 85),
                ("cd", "c", 80),
                ("cd", "d", 80),
                ("abcd", "abc", 80),
                ("abcd", "cd", 80),
            },
        ),
        # Multiple components test case
        (
            {
                "component": [1, 1, 2, 2],
                "left_id": ["a", "b", "x", "y"],
                "right_id": ["b", "c", "y", "z"],
                "probability": [90, 85, 95, 92],
            },
            {
                ("xy", "x", 95),
                ("xy", "y", 95),
                ("yz", "y", 92),
                ("yz", "z", 92),
                ("xyz", "xy", 92),
                ("xyz", "yz", 92),
                ("ab", "a", 90),
                ("ab", "b", 90),
                ("bc", "b", 85),
                ("bc", "c", 85),
                ("abc", "ab", 85),
                ("abc", "bc", 85),
            },
        ),
    ],
    ids=["single_component", "multiple_components"],
)
def test_hierarchical_clusters(input_data, expected_hierarchy):
    # Create input table
    probabilities = pa.table(
        input_data,
        schema=pa.schema(
            [
                ("component", pa.uint64()),
                ("left_id", pa.string()),
                ("right_id", pa.string()),
                ("probability", pa.uint8()),
            ]
        ),
    )

    # Create expected output table
    parents, children, probs = zip(*expected_hierarchy, strict=False)
    expected = (
        pa.table([parents, children, probs], names=["parent", "child", "probability"])
        .cast(
            pa.schema(
                [
                    ("parent", pa.string()),
                    ("child", pa.string()),
                    ("probability", pa.uint8()),
                ]
            )
        )
        .sort_by(
            [
                ("probability", "descending"),
                ("parent", "ascending"),
                ("child", "ascending"),
            ]
        )
    )

    # Run and compare
    result = to_hierarchical_clusters(
        probabilities,
        dtype=pa.string,
        proc_func=component_to_hierarchy,
        hash_func=_combine_strings,
    )

    result = result.sort_by(
        [
            ("probability", "descending"),
            ("parent", "ascending"),
            ("child", "ascending"),
        ]
    )

    assert result.schema == expected.schema
    assert result.equals(expected)



from os import environ
from typing import Callable, Generator
from unittest.mock import patch

import pytest
from fastapi.testclient import TestClient
from pydantic import SecretStr

from matchbox.client._settings import settings as client_settings
from matchbox.server import app


@pytest.fixture(scope="function")
def env_setter() -> Generator[Callable[[str, str], None], None, None]:
    """Set temporary env variable and refresh client settings."""
    original_values = {}

    def setter(var_name: str, var_value: str):
        original_values[var_name] = environ.get(var_name)

        environ[var_name] = var_value
        client_settings.__init__()

    yield setter

    for var_name, original_value in original_values.items():
        if original_value is None:
            del environ[var_name]
        else:
            environ[var_name] = original_value
    client_settings.__init__()


@pytest.fixture(scope="function")
def test_client() -> Generator[TestClient, None, None]:
    """Return a configured TestClient with patched backend and settings."""
    with (
        patch("matchbox.server.api.routes.settings") as mock_settings,
        patch("matchbox.server.api.routes.backend") as _,
    ):
        mock_settings.api_key = SecretStr("test-api-key")
        client = TestClient(app, headers={"X-API-Key": "test-api-key"})
        yield client



"""API routes for the Matchbox server."""

import logging
import sys
from contextlib import asynccontextmanager
from importlib.metadata import version
from typing import Annotated, AsyncGenerator

from fastapi import (
    BackgroundTasks,
    Body,
    Depends,
    FastAPI,
    HTTPException,
    Query,
    Security,
    UploadFile,
    status,
)
from fastapi.responses import JSONResponse, Response
from fastapi.security import APIKeyHeader
from starlette.exceptions import HTTPException as StarletteHTTPException

from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendCountableType,
    BackendRetrievableType,
    BackendUploadType,
    CountResult,
    ModelAncestor,
    ModelMetadata,
    ModelOperationStatus,
    ModelOperationType,
    NotFoundError,
    OKMessage,
    UploadStatus,
)
from matchbox.common.exceptions import (
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionNotFoundError,
    MatchboxServerFileError,
    MatchboxSourceNotFoundError,
)
from matchbox.common.graph import ResolutionGraph
from matchbox.common.logging import ASIMFormatter
from matchbox.common.sources import Match, Source, SourceAddress
from matchbox.server.api.arrow import table_to_s3
from matchbox.server.api.cache import MetadataStore, process_upload
from matchbox.server.base import (
    MatchboxDBAdapter,
    MatchboxServerSettings,
    get_backend_settings,
    settings_to_backend,
)


class ParquetResponse(Response):
    """A response object for returning parquet data."""

    media_type = "application/octet-stream"


settings: MatchboxServerSettings | None = None
backend: MatchboxDBAdapter | None = None
metadata_store = MetadataStore(expiry_minutes=30)
api_key_header = APIKeyHeader(name="X-API-Key")


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    """Context manager for the FastAPI lifespan events."""
    # Set up the backend
    global settings
    global backend

    SettingsClass = get_backend_settings(MatchboxServerSettings().backend_type)
    settings = SettingsClass()
    backend = settings_to_backend(settings)

    # Define common formatter
    formatter = ASIMFormatter()

    # Configure loggers with the same handler and formatter
    loggers_to_configure = [
        "matchbox",
        "uvicorn",
        "uvicorn.access",
        "uvicorn.error",
        "uvicorn.asgi",
        "fastapi",
    ]

    for logger_name in loggers_to_configure:
        # Configure handler
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(backend.settings.log_level)
        handler.setFormatter(formatter)

        logger = logging.getLogger(logger_name)
        logger.setLevel(backend.settings.log_level)
        # Remove any existing handlers to avoid duplicates
        if logger.handlers:
            logger.handlers.clear()
        logger.addHandler(handler)

    # Set SQLAlchemy loggers
    for sql_logger in ["sqlalchemy", "sqlalchemy.engine"]:
        logging.getLogger(sql_logger).setLevel("WARNING")

    yield

    del settings
    del backend


app = FastAPI(
    title="matchbox API",
    version=version("matchbox_db"),
    lifespan=lifespan,
)


@app.exception_handler(StarletteHTTPException)
async def http_exception_handler(request, exc):
    """Overwrite the default JSON schema for an `HTTPException`."""
    return JSONResponse(content=exc.detail, status_code=exc.status_code)


def validate_api_key(api_key: str = Security(api_key_header)) -> None:
    """Validate client API Key against settings."""
    if not settings.api_key:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="API Key missing in server configuration.",
            headers={"WWW-Authenticate": "X-API-Key"},
        )

    if not api_key:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="API Key required but not provided.",
            headers={"WWW-Authenticate": "X-API-Key"},
        )
    elif api_key != settings.api_key.get_secret_value():
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="API Key invalid.",
            headers={"WWW-Authenticate": "X-API-Key"},
        )


# General


@app.get("/health")
async def healthcheck() -> OKMessage:
    """Perform a health check and return the status."""
    return OKMessage()


@app.post(
    "/upload/{upload_id}",
    responses={
        400: {"model": UploadStatus, **UploadStatus.status_400_examples()},
    },
    status_code=status.HTTP_202_ACCEPTED,
    dependencies=[Depends(validate_api_key)],
)
async def upload_file(
    background_tasks: BackgroundTasks,
    upload_id: str,
    file: UploadFile,
) -> UploadStatus:
    """Upload and process a file based on metadata type.

    The file is uploaded to S3 and then processed in a background task.
    Status can be checked using the /upload/{upload_id}/status endpoint.

    Raises HTTP 400 if:

    * Upload ID not found or expired (entries expire after 30 minutes of inactivity)
    * Upload is already being processed
    * Uploaded data doesn't match the metadata schema
    """
    # Get and validate cache entry
    source_cache = metadata_store.get(cache_id=upload_id)
    if not source_cache:
        raise HTTPException(
            status_code=400,
            detail=UploadStatus(
                id=upload_id,
                status="failed",
                details=(
                    "Upload ID not found or expired. Entries expire after 30 minutes "
                    "of inactivity, including failed processes."
                ),
            ).model_dump(),
        )

    # Check if already processing
    if source_cache.status.status != "awaiting_upload":
        raise HTTPException(
            status_code=400,
            detail=source_cache.status.model_dump(),
        )

    # Upload to S3
    client = backend.settings.datastore.get_client()
    bucket = backend.settings.datastore.cache_bucket_name
    key = f"{upload_id}.parquet"

    try:
        await table_to_s3(
            client=client,
            bucket=bucket,
            key=key,
            file=file,
            expected_schema=source_cache.upload_type.schema,
        )
    except MatchboxServerFileError as e:
        metadata_store.update_status(upload_id, "failed", details=str(e))
        raise HTTPException(
            status_code=400,
            detail=source_cache.status.model_dump(),
        ) from e

    metadata_store.update_status(upload_id, "queued")

    # Start background processing
    background_tasks.add_task(
        process_upload,
        backend=backend,
        upload_id=upload_id,
        bucket=bucket,
        key=key,
        metadata_store=metadata_store,
    )

    source_cache = metadata_store.get(upload_id)

    # Check for error in async task
    if source_cache.status.status == "failed":
        raise HTTPException(
            status_code=400,
            detail=source_cache.status.model_dump(),
        )
    else:
        return source_cache.status


@app.get(
    "/upload/{upload_id}/status",
    responses={
        400: {"model": UploadStatus, **UploadStatus.status_400_examples()},
    },
    status_code=status.HTTP_200_OK,
)
async def get_upload_status(
    upload_id: str,
) -> UploadStatus:
    """Get the status of an upload process.

    Returns the current status of the upload and processing task.

    Raises HTTP 400 if:
    * Upload ID not found or expired (entries expire after 30 minutes of inactivity)
    """
    source_cache = metadata_store.get(cache_id=upload_id)
    if not source_cache:
        raise HTTPException(
            status_code=400,
            detail=UploadStatus(
                id=upload_id,
                status="failed",
                details=(
                    "Upload ID not found or expired. Entries expire after 30 minutes "
                    "of inactivity, including failed processes."
                ),
                entity=BackendUploadType.INDEX,
            ).model_dump(),
        )

    return source_cache.status


# Retrieval


@app.get(
    "/query",
    responses={404: {"model": NotFoundError}},
)
def query(
    full_name: str,
    warehouse_hash_b64: str,
    resolution_name: str | None = None,
    threshold: int | None = None,
    limit: int | None = None,
) -> ParquetResponse:
    """Query Matchbox for matches based on a source address."""
    source_address = SourceAddress(
        full_name=full_name, warehouse_hash=warehouse_hash_b64
    )
    try:
        res = backend.query(
            source_address=source_address,
            resolution_name=resolution_name,
            threshold=threshold,
            limit=limit,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e
    except MatchboxSourceNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.SOURCE
            ).model_dump(),
        ) from e

    buffer = table_to_buffer(res)
    return ParquetResponse(buffer.getvalue())


@app.get(
    "/match",
    responses={404: {"model": NotFoundError}},
)
def match(
    target_full_names: Annotated[list[str], Query()],
    target_warehouse_hashes_b64: Annotated[list[str], Query()],
    source_full_name: str,
    source_warehouse_hash_b64: str,
    source_pk: str,
    resolution_name: str,
    threshold: int | None = None,
) -> list[Match]:
    """Match a source primary key against a list of target addresses."""
    targets = [
        SourceAddress(full_name=n, warehouse_hash=w)
        for n, w in zip(target_full_names, target_warehouse_hashes_b64, strict=True)
    ]
    source = SourceAddress(
        full_name=source_full_name, warehouse_hash=source_warehouse_hash_b64
    )
    try:
        res = backend.match(
            source_pk=source_pk,
            source=source,
            targets=targets,
            resolution_name=resolution_name,
            threshold=threshold,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e
    except MatchboxSourceNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.SOURCE
            ).model_dump(),
        ) from e

    return res


# Data management


@app.post(
    "/sources",
    status_code=status.HTTP_202_ACCEPTED,
    dependencies=[Depends(validate_api_key)],
)
async def add_source(source: Source) -> UploadStatus:
    """Create an upload and insert task for indexed source data."""
    upload_id = metadata_store.cache_source(metadata=source)
    return metadata_store.get(cache_id=upload_id).status


@app.get(
    "/sources/{warehouse_hash_b64}/{full_name}",
    responses={404: {"model": NotFoundError}},
)
async def get_source(
    warehouse_hash_b64: str,
    full_name: str,
) -> Source:
    """Get a source from the backend."""
    address = SourceAddress(full_name=full_name, warehouse_hash=warehouse_hash_b64)
    try:
        return backend.get_source(address)
    except MatchboxSourceNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.SOURCE
            ).model_dump(),
        ) from e


@app.get(
    "/sources",
    responses={404: {"model": NotFoundError}},
)
async def get_resolution_sources(
    resolution_name: str,
) -> list[Source]:
    """Get all sources in scope for a resolution."""
    try:
        return backend.get_resolution_sources(resolution_name=resolution_name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e


@app.get("/report/resolutions")
async def get_resolutions() -> ResolutionGraph:
    """Get the resolution graph."""
    return backend.get_resolution_graph()


# Model management


@app.post(
    "/models",
    responses={
        500: {
            "model": ModelOperationStatus,
            **ModelOperationStatus.status_500_examples(),
        },
    },
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(validate_api_key)],
)
async def insert_model(model: ModelMetadata) -> ModelOperationStatus:
    """Insert a model into the backend."""
    try:
        backend.insert_model(model)
        return ModelOperationStatus(
            success=True,
            model_name=model.name,
            operation=ModelOperationType.INSERT,
        )
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ModelOperationStatus(
                success=False,
                model_name=model.name,
                operation=ModelOperationType.INSERT,
                details=str(e),
            ).model_dump(),
        ) from e


@app.get(
    "/models/{name}",
    responses={404: {"model": NotFoundError}},
)
async def get_model(name: str) -> ModelMetadata:
    """Get a model from the backend."""
    try:
        return backend.get_model(model=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e


@app.post(
    "/models/{name}/results",
    responses={404: {"model": NotFoundError}},
    status_code=status.HTTP_202_ACCEPTED,
    dependencies=[Depends(validate_api_key)],
)
async def set_results(name: str) -> UploadStatus:
    """Create an upload task for model results."""
    try:
        metadata = backend.get_model(model=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e

    upload_id = metadata_store.cache_model(metadata=metadata)
    return metadata_store.get(cache_id=upload_id).status


@app.get(
    "/models/{name}/results",
    responses={404: {"model": NotFoundError}},
)
async def get_results(name: str) -> ParquetResponse:
    """Download results for a model as a parquet file."""
    try:
        res = backend.get_model_results(model=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e

    buffer = table_to_buffer(res)
    return ParquetResponse(buffer.getvalue())


@app.patch(
    "/models/{name}/truth",
    responses={
        404: {"model": NotFoundError},
        500: {
            "model": ModelOperationStatus,
            **ModelOperationStatus.status_500_examples(),
        },
    },
    dependencies=[Depends(validate_api_key)],
)
async def set_truth(
    name: str,
    truth: Annotated[int, Body(ge=0, le=100)],
) -> ModelOperationStatus:
    """Set truth data for a model."""
    try:
        backend.set_model_truth(model=name, truth=truth)
        return ModelOperationStatus(
            success=True,
            model_name=name,
            operation=ModelOperationType.UPDATE_TRUTH,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ModelOperationStatus(
                success=False,
                model_name=name,
                operation=ModelOperationType.UPDATE_TRUTH,
                details=str(e),
            ).model_dump(),
        ) from e


@app.get(
    "/models/{name}/truth",
    responses={404: {"model": NotFoundError}},
)
async def get_truth(name: str) -> float:
    """Get truth data for a model."""
    try:
        return backend.get_model_truth(model=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e


@app.get(
    "/models/{name}/ancestors",
    responses={404: {"model": NotFoundError}},
)
async def get_ancestors(name: str) -> list[ModelAncestor]:
    """Get the ancestors for a model."""
    try:
        return backend.get_model_ancestors(model=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e


@app.patch(
    "/models/{name}/ancestors_cache",
    responses={
        404: {"model": NotFoundError},
        500: {
            "model": ModelOperationStatus,
            **ModelOperationStatus.status_500_examples(),
        },
    },
    dependencies=[Depends(validate_api_key)],
)
async def set_ancestors_cache(
    name: str,
    ancestors: list[ModelAncestor],
):
    """Update the cached ancestors for a model."""
    try:
        backend.set_model_ancestors_cache(model=name, ancestors_cache=ancestors)
        return ModelOperationStatus(
            success=True,
            model_name=name,
            operation=ModelOperationType.UPDATE_ANCESTOR_CACHE,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ModelOperationStatus(
                success=False,
                model_name=name,
                operation=ModelOperationType.UPDATE_ANCESTOR_CACHE,
                details=str(e),
            ).model_dump(),
        ) from e


@app.get(
    "/models/{name}/ancestors_cache",
    responses={404: {"model": NotFoundError}},
)
async def get_ancestors_cache(name: str) -> list[ModelAncestor]:
    """Get the cached ancestors for a model."""
    try:
        return backend.get_model_ancestors_cache(model=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e


@app.delete(
    "/models/{name}",
    responses={
        404: {"model": NotFoundError},
        409: {
            "model": ModelOperationStatus,
            **ModelOperationStatus.status_409_examples(),
        },
    },
    dependencies=[Depends(validate_api_key)],
)
async def delete_model(
    name: str,
    certain: Annotated[
        bool, Query(description="Confirm deletion of the model")
    ] = False,
) -> ModelOperationStatus:
    """Delete a model from the backend."""
    try:
        backend.delete_model(model=name, certain=certain)
        return ModelOperationStatus(
            success=True,
            model_name=name,
            operation=ModelOperationType.DELETE,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendRetrievableType.RESOLUTION
            ).model_dump(),
        ) from e
    except MatchboxDeletionNotConfirmed as e:
        raise HTTPException(
            status_code=409,
            detail=ModelOperationStatus(
                success=False,
                model_name=name,
                operation=ModelOperationType.DELETE,
                details=str(e),
            ).model_dump(),
        ) from e


# Admin


@app.get("/database/count")
async def count_backend_items(
    entity: BackendCountableType | None = None,
) -> CountResult:
    """Count the number of various entities in the backend."""

    def get_count(e: BackendCountableType) -> int:
        return getattr(backend, str(e)).count()

    if entity is not None:
        return CountResult(entities={str(entity): get_count(entity)})
    else:
        res = {str(e): get_count(e) for e in BackendCountableType}
        return CountResult(entities=res)


@app.delete(
    "/database",
    responses={409: {"model": str}},
    dependencies=[Depends(validate_api_key)],
)
async def clear_database(
    certain: Annotated[
        bool,
        Query(
            description=(
                "Confirm deletion of all data in the database whilst retaining tables"
            )
        ),
    ] = False,
) -> OKMessage:
    """Delete all data from the backend whilst retaining tables."""
    try:
        backend.clear(certain=certain)
        return OKMessage()
    except MatchboxDeletionNotConfirmed as e:
        raise HTTPException(
            status_code=409,
            detail=str(e),
        ) from e
